// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
#include <ntddk.h>
#include "filter.h"

u_long ntohl (u_long netlong)
{
	u_long result = 0;
	((char *)&result)[0] = ((char *)&netlong)[3];
	((char *)&result)[1] = ((char *)&netlong)[2];
	((char *)&result)[2] = ((char *)&netlong)[1];
	((char *)&result)[3] = ((char *)&netlong)[0];
	return result;
}

u_short ntohs (u_short netshort)
{
	u_short result = 0;
	((char *)&result)[0] = ((char *)&netshort)[1];
	((char *)&result)[1] = ((char *)&netshort)[0];
	return result;
}

// quick filter (TEST ONLY)
int quick_filter(const struct sockaddr *name_from, const struct sockaddr *name_to,
    int namelen, int proto, int direction, cont_flt_t *flt_in, cont_flt_t *flt_out)
{
    const struct sockaddr_in *from, *to;
	char *p;

	// default filters: None
	if (flt_in) *flt_in = 0;
	if (flt_out) *flt_out = 0;

	// not IP
    if (namelen != sizeof(struct sockaddr_in) ||
        name_from->sa_family != AF_INET ||
        name_to->sa_family != AF_INET)
    {
        return FILTER_DENY;
    }

    from = (const struct sockaddr_in *)name_from;
    to = (const struct sockaddr_in *)name_to;

	// quick filter
	switch (proto) {
	case IPPROTO_TCP:
		// TCP
		switch (direction) {
    
		case DIRECTION_IN:		/* incoming connections */
			
			// deny incoming port 21 connections
			if (ntohs(to->sin_port) == 21) return FILTER_DENY;

	        // content filter for incoming data for incoming port 80 connections
			if (ntohs(to->sin_port) == 80) {
				*flt_in = (cont_flt_t)1; // test filter
				*flt_out = (cont_flt_t)1; // test filter
				return FILTER_CONTENT;
			}

			return FILTER_ALLOW;
    
		case DIRECTION_OUT:		/* outgoing connections */
			
			// deny outgoing port 22 connections
			if (ntohs(to->sin_port) == 22) return FILTER_DENY;

	        // content filter for outgoing & incoming data for outgoing port 80 connections
			if (ntohs(to->sin_port) == 80) {
				*flt_in = (cont_flt_t)1; // test filter
				*flt_out = (cont_flt_t)1; // test filter
				return FILTER_CONTENT;
			}

			return FILTER_ALLOW;

		}
		break;

	case IPPROTO_UDP:
		// UDP
		switch (direction) {

			case DIRECTION_IN:	/* incoming datagrams */
				return FILTER_ALLOW;

			case DIRECTION_OUT:	/* outgoing datagrams */
				// filter data
				*flt_out = (cont_flt_t)1;
				return FILTER_CONTENT;

		}
		break;

	case IPPROTO_IP:
		// IP
		return FILTER_ALLOW; /* allow raw ip */
	}
	
	return FILTER_DENY;
}


// content filter (TEST ONLY)
int content_filter(cont_flt_t flt,
	const char *data, u_long data_size, u_long data_offset,
	char **new_data, u_long *new_data_size, alloc_fn_t *alloc_fn, int is_oob)
{
	char *p;

	if (flt == CONTFLT_ALLOWALL) return FILTER_ALLOW;
	if (flt == CONTFLT_DENYALL) return FILTER_DENY;

	// our test filter
	
	if (flt != (cont_flt_t)1) return FILTER_DENY;

	// if data contains 'SHIT' or 'shit'  word - drop connection (TCP) or datagram (UDP)
	if (memstr(data, data_size, "SHIT", 4) || memstr(data, data_size, "shit", 4)) {
		KdPrint(("[tdi_flt] content_filter: FILTER_DENY sequence detected\n"));
		return FILTER_DENY;
	}

	// if data contains 'FUCK' or 'fuck' word - replace with 'F_uc_K'
	p = memstr(data, data_size, "FUCK", 4);
	if (!p) p = memstr(data, data_size, "fuck", 4);

	if (p) {
		char *nd;
		ULONG offset = p - data;
		
		KdPrint(("[tdi_flt] content_filter: FILTER_CONTENT sequence detected\n"));

		if (!new_data || !new_data_size || !alloc_fn)
			return FILTER_DENY; // can't replace

		nd = (char *)alloc_fn(data_size + (6 - 4));
		if (!nd) return FILTER_DENY;

		memcpy(nd, data, offset);
		memcpy(nd + offset, "F_uc_K", 6);
		memcpy(nd + offset + 6, p + 4, data_size - offset - 4);

		*new_data = nd;
		*new_data_size = data_size + (6 - 4);
	}

	return FILTER_ALLOW;
}

// very slow and unoptimized subsequence finder
char *memstr(const char *data, u_long data_size, const char *key, u_long key_size)
{
	u_long i;
	char first_c;

	if (key_size == 0) return NULL;
	if (key_size > data_size) return NULL;
	if (key_size == data_size) return memcmp(data, key, key_size) == 0 ? (char *)data : NULL;

	first_c = *key;

	for (i = 0; i < data_size - key_size; i++, data++) {
		if (*data == first_c && memcmp(data, key, key_size) == 0) return (char *)data;
	}

	return NULL;
}
