// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
#include <ntddk.h>
#include <tdikrnl.h>
#include "filter.h"
#include "memtrack.h"
#include "obj_tbl.h"

#ifndef IoSkipCurrentIrpStackLocation
#	define IoSkipCurrentIrpStackLocation( Irp ) \
	    (Irp)->CurrentLocation++; \
	    (Irp)->Tail.Overlay.CurrentStackLocation++;
#endif

/*--- types ---*/

/* completion info */
typedef struct {
    PIO_COMPLETION_ROUTINE cr;
	PVOID context;
} TDI_COMPLETE_NFO;

/* what filter driver would do? */
/* tdi ioctl dispatcher function */
typedef int tdi_ioctl_fn_t(PIRP, PIO_STACK_LOCATION, TDI_COMPLETE_NFO *);

/*--- prototypes ---*/

static void *tdi_alloc(u_long size);
static char *get_mdl_data(PMDL mdl_data, ULONG *size, int *b_free);
static ULONG write_mdl(PMDL mdl_data, const char *data, ULONG size);

static VOID OnUnload(IN PDRIVER_OBJECT DriverObject);
static NTSTATUS DeviceDispatch(IN PDEVICE_OBJECT DeviceObject, IN PIRP irp);

static NTSTATUS
tdi_send_irp_to_old_driver(IN PDEVICE_OBJECT dev, IN PIRP irp, TDI_COMPLETE_NFO *tcn);

NTSTATUS
tdi_complete(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context);

static tdi_ioctl_fn_t
	tdi_associate_address,
	tdi_connect,
	tdi_disassociate_address,
	tdi_receive,
	tdi_receive_datagram,
	tdi_send,
	tdi_send_datagram,
	tdi_set_event_handler;

static tdi_ioctl_fn_t tdi_create, tdi_cleanup;

static NTSTATUS
tdi_event_connect(
    IN PVOID TdiEventContext,
    IN LONG RemoteAddressLength,
    IN PVOID RemoteAddress,
    IN LONG UserDataLength,
    IN PVOID UserData,
    IN LONG OptionsLength,
    IN PVOID Options,
    OUT CONNECTION_CONTEXT *ConnectionContext,
    OUT PIRP *AcceptIrp);

static NTSTATUS
tdi_event_disconnect(
    IN PVOID TdiEventContext,
    IN CONNECTION_CONTEXT ConnectionContext,
    IN LONG DisconnectDataLength,
    IN PVOID DisconnectData,
    IN LONG DisconnectInformationLength,
    IN PVOID DisconnectInformation,
    IN ULONG DisconnectFlags);

static NTSTATUS
tdi_event_receive(
    IN PVOID TdiEventContext,
    IN CONNECTION_CONTEXT ConnectionContext,
    IN ULONG ReceiveFlags,
    IN ULONG BytesIndicated,
    IN ULONG BytesAvailable,
    OUT ULONG *BytesTaken,
    IN PVOID Tsdu,
    OUT PIRP *IoRequestPacket);

static NTSTATUS
tdi_event_receive_datagram(
    IN PVOID TdiEventContext,
    IN LONG SourceAddressLength,
    IN PVOID SourceAddress,
    IN LONG OptionsLength,
    IN PVOID Options,
    IN ULONG ReceiveDatagramFlags,
    IN ULONG BytesIndicated,
    IN ULONG BytesAvailable,
    OUT ULONG *BytesTaken,
    IN PVOID Tsdu,
    OUT PIRP *IoRequestPacket);

static NTSTATUS
tdi_event_receive_expedited(
    IN PVOID TdiEventContext,
    IN CONNECTION_CONTEXT ConnectionContext,
    IN ULONG ReceiveFlags,
    IN ULONG BytesIndicated,
    IN ULONG BytesAvailable,
    OUT ULONG *BytesTaken,
    IN PVOID Tsdu,
    OUT PIRP *IoRequestPacket);

static NTSTATUS
tdi_event_chained_receive(
    IN PVOID TdiEventContext,
    IN CONNECTION_CONTEXT ConnectionContext,
    IN ULONG ReceiveFlags,
    IN ULONG ReceiveLength,
    IN ULONG StartingOffset,
    IN PMDL  Tsdu,
    IN PVOID TsduDescriptor);

static NTSTATUS
tdi_event_chained_receive_datagram(
    IN PVOID TdiEventContext,
    IN LONG SourceAddressLength,
    IN PVOID SourceAddress,
    IN LONG OptionsLength,
    IN PVOID Options,
    IN ULONG ReceiveDatagramFlags,
    IN ULONG ReceiveDatagramLength,
    IN ULONG StartingOffset,
    IN PMDL  Tsdu,
    IN PVOID TsduDescriptor);

static NTSTATUS
tdi_event_chained_receive_expedited(
    IN PVOID TdiEventContext,
    IN CONNECTION_CONTEXT ConnectionContext,
    IN ULONG ReceiveFlags,
    IN ULONG ReceiveLength,
    IN ULONG StartingOffset,
    IN PMDL  Tsdu,
    IN PVOID TsduDescriptor);

/*--- completion routines and their contexts */

typedef struct {
    PIO_COMPLETION_ROUTINE old_cr;
    PVOID old_context;
    PIO_COMPLETION_ROUTINE new_cr;
	PVOID new_context;
	PFILE_OBJECT fileobj;
} TDI_COMPLETION_CTX;

static NTSTATUS
tdi_completion(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context);

static NTSTATUS
tdi_create_addrobj_complete(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context);

typedef struct {
    MDL *old_mdl;
    char *buf;
} TDI_SEND_CONTEXT;

static NTSTATUS
tdi_send_complete(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context);

static NTSTATUS
tdi_receive_complete(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context);

static NTSTATUS
tdi_receive_datagram_complete(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context);

/*--- globals ---*/

PDEVICE_OBJECT g_tcpfltobj = NULL, g_tcpoldobj = NULL,
	g_udpfltobj = NULL, g_udpoldobj = NULL,
	g_ipfltobj = NULL, g_ipoldobj = NULL;

static struct {
	UCHAR MinorFunction;
	tdi_ioctl_fn_t *fn;
} g_tdi_ioctls[] = {
	{TDI_ASSOCIATE_ADDRESS, tdi_associate_address},
	{TDI_CONNECT, tdi_connect},
	{TDI_DISASSOCIATE_ADDRESS, tdi_disassociate_address},
	{TDI_RECEIVE, tdi_receive},
	{TDI_RECEIVE_DATAGRAM, tdi_receive_datagram},
	{TDI_SEND, tdi_send},
	{TDI_SEND_DATAGRAM, tdi_send_datagram},
	{TDI_SET_EVENT_HANDLER, tdi_set_event_handler}
};

static struct {
	LONG event;
	PVOID handler;
} g_tdi_event_handlers[] = {
	{TDI_EVENT_DISCONNECT, (PVOID)tdi_event_disconnect},
	{TDI_EVENT_CONNECT, (PVOID)tdi_event_connect},
	{TDI_EVENT_RECEIVE, (PVOID)tdi_event_receive},
	{TDI_EVENT_CHAINED_RECEIVE, (PVOID)tdi_event_chained_receive},
	{TDI_EVENT_RECEIVE_EXPEDITED, (PVOID)tdi_event_receive_expedited},
	{TDI_EVENT_CHAINED_RECEIVE_EXPEDITED, (PVOID)tdi_event_chained_receive_expedited},
	{TDI_EVENT_RECEIVE_DATAGRAM, (PVOID)tdi_event_receive_datagram},
	{TDI_EVENT_CHAINED_RECEIVE_DATAGRAM, (PVOID)tdi_event_chained_receive_datagram}
};

/* initialization */
NTSTATUS
DriverEntry(IN PDRIVER_OBJECT theDriverObject,
            IN PUNICODE_STRING theRegistryPath)
{
    NTSTATUS status = STATUS_SUCCESS;
    UNICODE_STRING devname;
	int i;

	memtrack_init();

	status = ot_init();
	if (status) {
		KdPrint(("[tdi_flt] DriverEntry: ot_init: 0x%x\n", status));
		goto done;
	}

	for (i = 0; i < IRP_MJ_MAXIMUM_FUNCTION; i++)
		theDriverObject->MajorFunction[i] = DeviceDispatch;

	// register UnLoad procedure
	theDriverObject->DriverUnload  = OnUnload;

	/*
     * the code below intialize TCP TDI Filter Driver
     */
	status = IoCreateDevice(theDriverObject,
							0,
							NULL,
							FILE_DEVICE_UNKNOWN,
							0,
							TRUE,
							&g_tcpfltobj);
	if (status) {
		KdPrint(("[tdi_flt] DriverEntry: IoCreateDevice(Tcp): 0x%x\n", status));
		goto done;
	}

	g_tcpfltobj->Flags |= DO_DIRECT_IO;

	/* attach to original TCP driver */

	RtlInitUnicodeString(&devname, L"\\Device\\Tcp");
	status = IoAttachDevice(g_tcpfltobj, &devname, &g_tcpoldobj);
	if (status) {
		KdPrint(("[tdi_flt] DriverEntry: IoAttachDevice(Tcp): 0x%x\n", status));
		goto done;
	}

	KdPrint(("[tdi_flt] DriverEntry: TCP fileobj: 0x%x\n",g_tcpfltobj ));

	/*
     * the code below intialize UDP TDI Filter Driver
     */
	status = IoCreateDevice(theDriverObject,
							0,
							NULL,
							FILE_DEVICE_UNKNOWN,
							0,
							TRUE,
							&g_udpfltobj);
	if (status) {
		KdPrint(("[tdi_flt] DriverEntry: IoCreateDevice(Udp): 0x%x\n", status));
		goto done;
	}

	g_udpfltobj->Flags |= DO_DIRECT_IO;

	/* attach to original UDP driver */

	RtlInitUnicodeString(&devname, L"\\Device\\Udp");
	status = IoAttachDevice(g_udpfltobj, &devname, &g_udpoldobj);
	if (status) {
		KdPrint(("[tdi_flt] DriverEntry: IoAttachDevice(Udp): 0x%x\n", status));
		goto done;
	}

	KdPrint(("[tdi_flt] DriverEntry: UDP fileobj: 0x%x\n", g_udpfltobj));

	/*
     * the code below intialize IP TDI Filter Driver
     */
	status = IoCreateDevice(theDriverObject,
							0,
							NULL,
							FILE_DEVICE_UNKNOWN,
							0,
							TRUE,
							&g_ipfltobj);
	if (status) {
		KdPrint(("[tdi_flt] DriverEntry: IoCreateDevice(Ip): 0x%x\n", status));
		goto done;
	}

	g_ipfltobj->Flags |= DO_DIRECT_IO;

	/* attach to original IP driver */

	RtlInitUnicodeString(&devname, L"\\Device\\RawIp");
	status = IoAttachDevice(g_ipfltobj, &devname, &g_ipoldobj);
	if (status) {
		KdPrint(("[tdi_flt] DriverEntry: IoAttachDevice(Ip): 0x%x\n", status));
		goto done;
	}

	KdPrint(("[tdi_flt] DriverEntry: IP fileobj: 0x%x\n",g_tcpfltobj ));

	status = STATUS_SUCCESS;
done:
	if (status) {
		// cleanup
		OnUnload(theDriverObject);
	}

    return status;
}

/* deinitialization */
VOID
OnUnload(IN PDRIVER_OBJECT DriverObject)
{
	if (g_tcpoldobj) IoDetachDevice(g_tcpoldobj);
	if (g_tcpfltobj) IoDeleteDevice(g_tcpfltobj);

	if (g_udpoldobj) IoDetachDevice(g_udpoldobj);
	if (g_udpfltobj) IoDeleteDevice(g_udpfltobj);

	if (g_ipoldobj) IoDetachDevice(g_ipoldobj);
	if (g_ipfltobj) IoDeleteDevice(g_ipfltobj);

	ot_free();
	memtrack_free();
}

/* dispatch */
NTSTATUS
DeviceDispatch(IN PDEVICE_OBJECT DeviceObject, IN PIRP irp)
{
	PIO_STACK_LOCATION irps;

	if (!irp) {
		KdPrint(("[tdi_flt] DeviceDispatch: !irp\n"));
		return STATUS_SUCCESS;
	}
	
	irps = IoGetCurrentIrpStackLocation(irp);

	if (DeviceObject == g_tcpfltobj ||
		DeviceObject == g_udpfltobj ||
		DeviceObject == g_ipfltobj)
	{
		int fa = FILTER_ALLOW;
		TDI_COMPLETE_NFO tcn;
		NTSTATUS status;

		tcn.cr = NULL;
		irp->IoStatus.Status = STATUS_ACCESS_VIOLATION; // default for FILTER_DROP

		switch (irps->MajorFunction) {

		case IRP_MJ_CREATE:
			/* create fileobject */
			fa = tdi_create(irp, irps, &tcn);
			break;

		case IRP_MJ_DEVICE_CONTROL:
			KdPrint(("[tdi_flt] DeviceDispatch: IRP_MJ_DEVICE_CONTROL, minor 0x%x for 0x%08X\n",
				irps->MinorFunction, irps->FileObject));

			/* try to convert it to IRP_MJ_INTERNAL_DEVICE_CONTROL */
			status = TdiMapUserRequest(DeviceObject, irp, irps);
			if (status) {
				KdPrint(("[tdi_flt] DeviceDispatch: TdiMapUserRequest: 0x%x\n",
					status));
				break;
			}

			// no break! go to internal device control!
		
		case IRP_MJ_INTERNAL_DEVICE_CONTROL:
		{
			/* ioctl for tdi */
			int i;

			for (i = 0; i < sizeof(g_tdi_ioctls) / sizeof(*g_tdi_ioctls); i++)
				if (g_tdi_ioctls[i].MinorFunction == irps->MinorFunction) {
					fa = g_tdi_ioctls[i].fn(irp, irps, &tcn);
					break;
				}
	
			if (i >= sizeof(g_tdi_ioctls) / sizeof(*g_tdi_ioctls))
				KdPrint(("[tdi_flt] DeviceDispatch: IRP_MJ_INTERNAL_DEVICE_CONTROL, minor 0x%x for 0x%08X\n",
					irps->MinorFunction, irps->FileObject));

			break;
		}

		case IRP_MJ_CLEANUP:
			/* cleanup fileobject */
			fa = tdi_cleanup(irp, irps, &tcn);
			break;

		case IRP_MJ_CLOSE:
			KdPrint(("[tdi_flt] DeviceDispatch: IRP_MJ_CLOSE fileobj 0x%x\n", irps->FileObject));
			break;

		default:
			KdPrint(("[tdi_flt] DeviceDispatch: major 0x%x, minor 0x%x for 0x%08X\n",
				irps->MajorFunction, irps->MinorFunction, irps->FileObject));
		}

		if (fa == FILTER_ALLOW) {
			KdPrint(("[tdi_flt] DeviceDispatch: [ALLOW] major 0x%x, minor 0x%x for devobj 0x%x; fileobj 0x%x\n",
				irps->MajorFunction, irps->MinorFunction, DeviceObject, irps->FileObject));

			irp->IoStatus.Status = STATUS_SUCCESS;

			return tdi_send_irp_to_old_driver(DeviceObject, irp, &tcn);
		}
		else if (fa == FILTER_CONTENT) {
			KdPrint(("[tdi_flt] DeviceDispatch: [CONTENT]  major 0x%x, minor 0x%x for devobj 0x%x; fileobj 0x%x\n",
				irps->MajorFunction, irps->MinorFunction, DeviceObject, irps->FileObject));

			// no need to call filtered driver
			// content provided by us
			
			IoCompleteRequest (irp, IO_NO_INCREMENT);
			return irp->IoStatus.Status;
		}
	}
	else {
		// unknown device object
		KdPrint(("[tdi_flt] DeviceDispatch: unknown DeviceObject 0x%x\n", DeviceObject));
	}

	// DROP IT!
	KdPrint(("[tdi_flt] DeviceDispatch: [DROP!]  major 0x%x, minor 0x%x for devobj 0x%x; fileobj 0x%x\n",
		irps->MajorFunction, irps->MinorFunction, DeviceObject, irps->FileObject));
	
	IoCompleteRequest (irp, IO_NO_INCREMENT);

	return irp->IoStatus.Status;
}

/* send irp to old driver*/
NTSTATUS
tdi_send_irp_to_old_driver(IN PDEVICE_OBJECT dev, IN PIRP irp, TDI_COMPLETE_NFO *tcn)
{
	if (irp->CurrentLocation <= 1) {
		// we use _THIS_ way to send IRP to old driver to avoid NO_MORE_STACK_LOCATIONS
		IoSkipCurrentIrpStackLocation(irp);

		if (tcn && tcn->cr) {
			PIO_STACK_LOCATION irps = IoGetNextIrpStackLocation(irp);

			// save old completion routine and context
			TDI_COMPLETION_CTX *ctx = (TDI_COMPLETION_CTX *)malloc(sizeof(*ctx));
			if (!ctx) {
				KdPrint(("[tdi_flt] tdi_send_irp_to_old_driver: malloc\n"));
				irp->IoStatus.Status = STATUS_NO_MEMORY;
				IoCompleteRequest (irp, IO_NO_INCREMENT);
				return irp->IoStatus.Status;
			}

			ctx->old_cr = irps->CompletionRoutine;
			ctx->old_context = irps->Context;
			ctx->new_cr = tcn->cr;
			ctx->new_context = tcn->context;
			ctx->fileobj = irps->FileObject;

			IoSetCompletionRoutine(irp, tdi_completion, ctx, TRUE, TRUE, TRUE);
		}
	}
	else {
		PIO_STACK_LOCATION irps = IoGetCurrentIrpStackLocation(irp),
			next_irps = IoGetNextIrpStackLocation(irp);
		
		memcpy(next_irps, irps, sizeof(*irps));

		if (tcn && tcn->cr) {
			IoSetCompletionRoutine(irp, tcn->cr, tcn->context, TRUE, TRUE, TRUE);
		}
		else {
			IoSetCompletionRoutine(irp, tdi_complete, NULL, TRUE, TRUE, TRUE);
		}
	}

	if (dev == g_tcpfltobj) return IoCallDriver(g_tcpoldobj, irp);
	else if (dev == g_udpfltobj) return IoCallDriver(g_udpoldobj, irp);
	else if (dev == g_ipfltobj) return IoCallDriver(g_ipoldobj, irp);

	KdPrint(("[tdi_flt] tdi_send_irp_to_old_driver: Unknown DeviceObject 0x%x!\n", dev));
	return STATUS_UNSUCCESSFUL;
}

NTSTATUS
tdi_completion(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context)
{
	TDI_COMPLETION_CTX *ctx = (TDI_COMPLETION_CTX *)Context;
	NTSTATUS status;
	PIO_STACK_LOCATION irps;

	if (Irp->IoStatus.Status)
		KdPrint(("[tdi_flt] tdi_complete: status 0x%x\n", Irp->IoStatus.Status));

	Irp->CurrentLocation--;
	Irp->Tail.Overlay.CurrentStackLocation--;

	irps = IoGetCurrentIrpStackLocation(Irp);

	KdPrint(("[tdi_flt] tdi_completion: DeviceObject = 0x%x; FileObject = 0x%x\n",
		irps->DeviceObject, irps->FileObject));

	if (ctx->new_cr) {
		// restore fileobject
		irps->FileObject = ctx->fileobj;
		
		status = ctx->new_cr(irps->DeviceObject, Irp, ctx->new_context);
	}

	if (ctx->old_cr) {
		// restore routine and context
		irps->CompletionRoutine = ctx->old_cr;
		irps->Context = ctx->old_context;

		status = ctx->old_cr(irps->DeviceObject, Irp, ctx->old_context);
	}

	free(ctx);
	return status;
}

NTSTATUS
tdi_complete(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context)
{
	// generic completion
	
	if (Irp->IoStatus.Status)
		KdPrint(("[tdi_filter] tdi_complete: status 0x%x\n", Irp->IoStatus.Status));

	if (Irp->PendingReturned) IoMarkIrpPending(Irp);

	return STATUS_SUCCESS;
}

/*[--- TDI dispatchers ---]*/

int
tdi_create(PIRP irp, PIO_STACK_LOCATION irps, TDI_COMPLETE_NFO *tcn)
{
	NTSTATUS status;
	FILE_FULL_EA_INFORMATION *ea = (FILE_FULL_EA_INFORMATION *)irp->AssociatedIrp.SystemBuffer;

	if (ea) {
		if (ea->EaNameLength == TDI_TRANSPORT_ADDRESS_LENGTH &&
			memcmp(ea->EaName, TdiTransportAddress, TDI_TRANSPORT_ADDRESS_LENGTH) == 0)
		{
			KdPrint(("[tdi_flt] tdi_create: devobj 0x%x; addrobj 0x%x\n",
				irps->DeviceObject, irps->FileObject));

			status = ot_add_fileobj(irps->DeviceObject, irps->FileObject, FILEOBJ_ADDROBJ, NULL);
			if (status) {
				KdPrint(("[tdi_flt] tdi_create: ot_add_fileobj: 0x%x\n", status));
				return FILTER_DENY;
			}

			tcn->cr = tdi_create_addrobj_complete;
			tcn->context = NULL;
		}
		else if (ea->EaNameLength == TDI_CONNECTION_CONTEXT_LENGTH &&
			memcmp(ea->EaName, TdiConnectionContext, TDI_CONNECTION_CONTEXT_LENGTH) == 0)
		{
			CONNECTION_CONTEXT conn_ctx = *(CONNECTION_CONTEXT *)
				(ea->EaName + ea->EaNameLength + 1);

			KdPrint(("[tdi_flt] tdi_create: devobj 0x%x; connobj 0x%x; conn_ctx 0x%x\n",
				irps->DeviceObject, irps->FileObject, conn_ctx));

			status = ot_add_fileobj(irps->DeviceObject, irps->FileObject, FILEOBJ_CONNOBJ, conn_ctx);
			if (status) {
				KdPrint(("[tdi_flt] tdi_create: ot_add_fileobj: 0x%x\n", status));
				return FILTER_DENY;
			}
		}
	}
	else {
		KdPrint(("[tdi_flt] tdi_create: devobj 0x%x; Control Object: 0x%x\n",
			irps->DeviceObject, irps->FileObject));
	}

	return FILTER_ALLOW;
}

//----------------------------------------------------------------------------

NTSTATUS
tdi_create_addrobj_complete(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context)
{
	NTSTATUS status;
	PIO_STACK_LOCATION irps = IoGetCurrentIrpStackLocation(Irp);
	PIRP query_irp = NULL;
	KEVENT event;
	IO_STATUS_BLOCK isb;
	PMDL mdl;
	TDI_ADDRESS_INFO *tai = NULL;
	TA_ADDRESS *addr;
	ot_entry_t *ote = NULL;
	KIRQL irql;

	KdPrint(("[tdi_flt] tdi_create_addrobj_complete: devobj 0x%x; addrobj 0x%x\n",
		DeviceObject, irps->FileObject));

	if (Irp->IoStatus.Status != STATUS_SUCCESS) {
		KdPrint(("[tdi_flt] tdi_create_addrobj_complete: status 0x%x\n", Irp->IoStatus.Status));
		goto done;
	}

	// query addrobj address:port

	KeInitializeEvent(&event, SynchronizationEvent, FALSE);

	query_irp = TdiBuildInternalDeviceControlIrp(TDI_QUERY_INFORMATION, DeviceObject,
		irps->FileObject, &event, &isb);
	if (!query_irp) {
		KdPrint(("[tdi_flt] tdi_create_addrobj_complete: TdiBuildInternalDeviceControlIrp\n"));
		status = STATUS_UNSUCCESSFUL;
		goto done;
	}

	tai = (TDI_ADDRESS_INFO *)
		malloc(sizeof(TDI_ADDRESS_INFO) + TDI_ADDRESS_MAX_LENGTH);
	if (!tai) {
		KdPrint(("[tdi_flt] tdi_create_addrobj_complete: malloc\n"));
		status = STATUS_NO_MEMORY;
		goto done;
	}

	mdl = IoAllocateMdl(tai, TDI_ADDRESS_INFO_MAX, FALSE, FALSE, NULL);
	if (!mdl) {
		KdPrint(("[tdi_flt] tdi_create_addrobj_complete: IoAllocateMdl\n"));
		status = STATUS_NO_MEMORY;
		goto done;
	}
	MmProbeAndLockPages(mdl, KernelMode, IoWriteAccess);

	TdiBuildQueryInformation(query_irp, DeviceObject, irps->FileObject, NULL, NULL,
		TDI_QUERY_ADDRESS_INFO, mdl);

	status = IoCallDriver(DeviceObject, query_irp);
	query_irp = NULL;
	mdl = NULL;

	if (status) {
		KdPrint(("[tdi_flt] tdi_create_addrobj_complete: IoCallDriver: 0x%x\n", status));
		goto done;
	}

	status = KeWaitForSingleObject(&event, Executive, KernelMode, FALSE, 0);
	if (status) {
		KdPrint(("[tdi_flt] tdi_create_addrobj_complete: KeWaitForSingleObject: 0x%x\n", status));
		goto done;
	}

	addr = tai->Address.Address;

	KdPrint(("[tdi_flt] tdi_create_addrobj_complete: address: %x:%u\n", 
		 ntohl(((TDI_ADDRESS_IP *)(addr->Address))->in_addr),
		 ntohs(((TDI_ADDRESS_IP *)(addr->Address))->sin_port)));

	// save address

	ote = ot_find_fileobj(irps->FileObject, &irql);
	if (!ote) {
		KdPrint(("[tdi_flt] tdi_create_addrobj_complete: ot_find_fileobj(0x%x)\n",
			irps->FileObject));
		status = STATUS_OBJECT_NAME_NOT_FOUND;
		goto done;
	}

	if (addr->AddressLength > sizeof(ote->local_addr)) {
		KdPrint(("[tdi_flt] tdi_create_addrobj_complete: address too long! (%u)\n",
			addr->AddressLength));
		status = STATUS_BUFFER_OVERFLOW;
		goto done;
	}
	memcpy(ote->local_addr, addr, addr->AddressLength);

	status = STATUS_SUCCESS;
done:
	if (ote) KeReleaseSpinLock(&g_ot_hash_guard, irql);
	if (mdl) IoFreeMdl(mdl);
	if (tai) free(tai);
	if (query_irp) IoFreeIrp(query_irp);

	Irp->IoStatus.Status = status;
	return tdi_complete(DeviceObject, Irp, Context);
}

//----------------------------------------------------------------------------

int
tdi_cleanup(PIRP irp, PIO_STACK_LOCATION irps, TDI_COMPLETE_NFO *tcn)
{
	NTSTATUS status;
	int type;
	CONNECTION_CONTEXT conn_ctx;
	
	status = ot_del_fileobj(irps->FileObject, &type);
	if (status) KdPrint(("[tdi_flt] tdi_cleanup: del_fileobj: 0x%x\n", status));
	else KdPrint(("[tdi_flt] tdi_cleanup: fileobj 0x%x, type %d\n", irps->FileObject, type));

	return FILTER_ALLOW;
}

//----------------------------------------------------------------------------

int
tdi_associate_address(PIRP irp, PIO_STACK_LOCATION irps, TDI_COMPLETE_NFO *tcn)
{
	HANDLE addr_handle = ((TDI_REQUEST_KERNEL_ASSOCIATE *)(&irps->Parameters))->AddressHandle;
	PFILE_OBJECT addrobj;
	NTSTATUS status;
	ot_entry_t *ote_conn = NULL;
	KIRQL irql;
	int result = FILTER_DENY;

	KdPrint(("[tdi_flt] tdi_associate_address: devobj 0x%x; connobj 0x%x\n",
		irps->DeviceObject, irps->FileObject));

	status = ObReferenceObjectByHandle(addr_handle, GENERIC_READ, NULL, KernelMode, &addrobj, NULL);
	if (status) {
		KdPrint(("[tdi_flt] tdi_associate_address: ObReferenceObjectByHandle: 0x%x\n", status));
		goto done;
	}

	KdPrint(("[tdi_flt] tdi_associate_address: connobj = 0x%x ---> addrobj = 0x%x\n",
		irps->FileObject, addrobj));

	// associate addrobj with connobj

	ote_conn = ot_find_fileobj(irps->FileObject, &irql);
	if (!ote_conn) {
		KdPrint(("[tdi_flt] tdi_associate_address: ot_find_fileobj(0x%x)\n", irps->FileObject));
		goto done;
	}
	ote_conn->associated_fileobj = addrobj;

	// add (conn_ctx, addrobj)->connobj

	status = ot_add_conn_ctx(addrobj, ote_conn->conn_ctx, irps->FileObject);
	if (status) {
		KdPrint(("[tdi_flt] tdi_associate_address: ot_add_conn_ctx: 0x%x\n", status));
		goto done;
	}

	result = FILTER_ALLOW;
done:
	if (ote_conn) KeReleaseSpinLock(&g_ot_hash_guard, irql);

	return result;
}

//----------------------------------------------------------------------------

int
tdi_connect(PIRP irp, PIO_STACK_LOCATION irps, TDI_COMPLETE_NFO *tcn)
{
	PTDI_REQUEST_KERNEL_CONNECT request = (PTDI_REQUEST_KERNEL_CONNECT)(&irps->Parameters);
	TA_ADDRESS *remote_addr = ((TRANSPORT_ADDRESS *)(request->RequestConnectionInformation->RemoteAddress))->Address;
	PFILE_OBJECT addrobj;
	NTSTATUS status;
	TA_ADDRESS *local_addr;
	int result = FILTER_DENY, ipproto;
	ot_entry_t *ote_conn = NULL, *ote_addr;
	KIRQL irql;
	cont_flt_t flt_in, flt_out;

	KdPrint(("[tdi_flt] tdi_connect: connobj 0x%x, address %x:%u\n",
		irps->FileObject, ntohl(((TDI_ADDRESS_IP *)(remote_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(remote_addr->Address))->sin_port)));

	if (irps->DeviceObject != g_tcpfltobj && irps->DeviceObject != g_udpfltobj) {
		// unknown device object!
		KdPrint(("[tdi_flt] tdi_connect: unknown DeviceObject 0x%x\n", irps->DeviceObject));
		goto done;
	}

	ote_conn = ot_find_fileobj(irps->FileObject, &irql);
	if (!ote_conn) {
		KdPrint(("[tdi_flt] tdi_connect: ot_find_fileobj(0x%x)\n", irps->FileObject));
		goto done;
	}

	if (irps->DeviceObject == g_tcpfltobj) {
		// get addrobj by connobj and get local address by it

		addrobj = ote_conn->associated_fileobj;
		if (!addrobj) {
			KdPrint(("[tdi_flt] tdi_connect: empty addrobj!\n"));
			goto done;
		}

		ote_addr = ot_find_fileobj(addrobj, NULL); // we're already in spinlock
		if (!ote_conn) {
			KdPrint(("[tdi_flt] tdi_connect: ot_find_fileobj(0x%x)\n", addrobj));
			goto done;
		}

		ipproto = IPPROTO_TCP;
	}
	else {
		KdPrint(("[tdi_flt] tdi_connect: connected UDP socket detected\n"));

		// for connected UDP sockets connobj and addrobj are the same
		addrobj= irps->FileObject;
		ote_addr = ote_conn;

		ipproto = IPPROTO_UDP;
	}

	local_addr = (TA_ADDRESS *)(ote_addr->local_addr);

	if (local_addr->AddressLength != remote_addr->AddressLength) {
		// what the ...
		KdPrint(("[tdi_flt] tdi_connect: different addr lengths! (%u != %u)\n",
			local_addr->AddressLength, remote_addr->AddressLength));
		goto done;
	}

	// associate remote address with connobj
	
	if (remote_addr->AddressLength > sizeof(ote_conn->remote_addr)) {
		KdPrint(("[tdi_flt] tdi_connect: address too long! (%u)\n", remote_addr->AddressLength));
		goto done;
	}
	memcpy(ote_conn->remote_addr, remote_addr, remote_addr->AddressLength);

	// associate local address with connobj

	if (local_addr->AddressLength > sizeof(ote_conn->local_addr)) {
		KdPrint(("[tdi_flt] tdi_connect: address to long! (%u)\n", local_addr->AddressLength));
		goto done;
	}
	memcpy(ote_conn->local_addr, local_addr, local_addr->AddressLength);

	// run filter

	KdPrint(("[tdi_flt] tdi_connect: %x:%u -> %x:%u (ipproto = %d)\n",
		ntohl(((TDI_ADDRESS_IP *)(local_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(local_addr->Address))->sin_port),
		ntohl(((TDI_ADDRESS_IP *)(remote_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(remote_addr->Address))->sin_port), ipproto));

	result = quick_filter((struct sockaddr *)(&local_addr->AddressType),
		(struct sockaddr *)(&remote_addr->AddressType), sizeof(struct sockaddr_in),
		ipproto, DIRECTION_OUT, &flt_in, &flt_out);
	
	KdPrint(("[tdi_flt] tdi_connect: quick_filter: %d\n", result));

	if (result == FILTER_CONTENT) {
		ote_conn->flt_in = flt_in;
		ote_conn->flt_out = flt_out;
		result = FILTER_ALLOW;
	}

done:
	if (ote_conn) KeReleaseSpinLock(&g_ot_hash_guard, irql);
	if (result != FILTER_ALLOW) irp->IoStatus.Status = STATUS_REMOTE_NOT_LISTENING; // <--??
	
	return result;
}

//----------------------------------------------------------------------------

int
tdi_disassociate_address(PIRP irp, PIO_STACK_LOCATION irps, TDI_COMPLETE_NFO *tcn)
{
	ot_entry_t *ote_conn = NULL;
	KIRQL irql;
	NTSTATUS status;

	KdPrint(("[tdi_flt] tdi_disassociate_address: connobj 0x%x\n", irps->FileObject));

	ote_conn = ot_find_fileobj(irps->FileObject, &irql);
	if (!ote_conn) {
		KdPrint(("[tdi_flt] tdi_disassociate_address: ot_find_fileobj(0x%x)\n", irps->FileObject));
		goto done;
	}

	status = ot_del_conn_ctx(ote_conn->associated_fileobj, ote_conn->conn_ctx);
	if (status) {
		KdPrint(("[tdi_flt] tdi_disassociate_address: ot_del_conn_ctx: 0x%x\n", status));
		goto done;
	}

done:
	if (ote_conn) KeReleaseSpinLock(&g_ot_hash_guard, irql);

	return FILTER_ALLOW; // always allow to do it
}

//----------------------------------------------------------------------------

int
tdi_receive(PIRP irp, PIO_STACK_LOCATION irps, TDI_COMPLETE_NFO *tcn)
{
	TDI_REQUEST_KERNEL_RECEIVE *param = (TDI_REQUEST_KERNEL_RECEIVE *)(&irps->Parameters);
	TA_ADDRESS *local_addr, *remote_addr;
	NTSTATUS status;
	ot_entry_t *ote = NULL;
	KIRQL irql;
	int result = FILTER_DENY;
	datapipe_t *pipe;
	ULONG data_size;

	if (irps->DeviceObject != g_tcpfltobj && irps->DeviceObject != g_udpfltobj) {
		// unknown device object!
		KdPrint(("[tdi_flt] tdi_receive: unknown DeviceObject 0x%x\n", irps->DeviceObject));
		goto done;
	}

	// get local and remote addresses of connection

	ote = ot_find_fileobj(irps->FileObject, &irql);
	if (!ote) {
		KdPrint(("[tdi_flt] tdi_receive: ot_find_fileobj(0x%x)\n", irps->FileObject));
		goto done;
	}

	KdPrint(("[tdi_flt] tdi_receive: connobj 0x%x (offset: %u; size: %u)\n", irps->FileObject,
		ote->in_offset, param->ReceiveLength));

	local_addr = (TA_ADDRESS *)(ote->local_addr);
	remote_addr = (TA_ADDRESS *)(ote->remote_addr);

	KdPrint(("[tdi_flt] tdi_receive: %x:%u -> %x:%u\n",
		ntohl(((TDI_ADDRESS_IP *)(remote_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(remote_addr->Address))->sin_port),
		ntohl(((TDI_ADDRESS_IP *)(local_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(local_addr->Address))->sin_port)));

	// try to read data from datapipe

	if (param->ReceiveFlags & TDI_RECEIVE_EXPEDITED) pipe = &ote->in_oob_pipe;
	else pipe = &ote->in_pipe;

	data_size = datapipe_peek(pipe);
	if (data_size) {
		// read data into MDL
		if (!irp->MdlAddress->Next) {
			// if unchained just read
			irp->IoStatus.Information = datapipe_get(pipe,
				MmGetSystemAddressForMdl(irp->MdlAddress), irp->MdlAddress->ByteCount);
		}
		else {
			// if chained
			char *data = (char *)malloc(data_size);
			data_size = datapipe_get(pipe, MmGetSystemAddressForMdl(irp->MdlAddress),
				irp->MdlAddress->ByteCount);
			irp->IoStatus.Information = write_mdl(irp->MdlAddress, data, data_size);
			free(data);
		}

		if (param->ReceiveFlags & TDI_RECEIVE_EXPEDITED)
			ote->in_oob_offset += irp->IoStatus.Information;
		else ote->in_offset += irp->IoStatus.Information;
		
		result = FILTER_CONTENT;
		goto done;
	}

	if (!ote->flt_in) {
		// don't filter - allow all

		if (param->ReceiveFlags & TDI_RECEIVE_EXPEDITED)
			ote->in_oob_offset += irp->IoStatus.Information;
		else ote->in_offset += irp->IoStatus.Information;
		
		result = FILTER_ALLOW;
		goto done;
	}

	// set content filter in completion
	tcn->cr = tdi_receive_complete;
	tcn->context = irps->FileObject;

	result = FILTER_ALLOW;
done:
	if (ote) KeReleaseSpinLock(&g_ot_hash_guard, irql);

	return result;
}

//----------------------------------------------------------------------------

NTSTATUS
tdi_receive_complete(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context)
{
	PIO_STACK_LOCATION irps = IoGetCurrentIrpStackLocation(Irp);
	TDI_REQUEST_KERNEL_RECEIVE *param = (TDI_REQUEST_KERNEL_RECEIVE *)(&irps->Parameters);
	PFILE_OBJECT connobj = (PFILE_OBJECT)Context; // can't use irps->FileObject
	ULONG new_data_size, data_size;
	char *data = NULL, *new_data = NULL;
	PMDL mdl;
	ot_entry_t *ote_conn = NULL;
	KIRQL irql;
	int result = FILTER_DENY, b_free_data = 0;
	NTSTATUS status;

	KdPrint(("[tdi_flt] tdi_receive_complete: connobj 0x%x; status 0x%x\n", connobj,
		Irp->IoStatus.Status));

	if (Irp->IoStatus.Status != STATUS_SUCCESS) {
		KdPrint(("[tdi_flt] tdi_receive_complete: status 0x%x\n", Irp->IoStatus.Status));
		goto done;
	}

	ote_conn = ot_find_fileobj(connobj, &irql);
	if (!ote_conn) {
		KdPrint(("[tdi_flt] tdi_receive_complete: ot_find_fileobj(0x%x)\n", connobj));
		goto done;
	}

	data = get_mdl_data(Irp->MdlAddress, &data_size, &b_free_data);
	if (!data) {
		KdPrint(("[tdi_flt] tdi_receive_complete: get_mdl_data\n"));
		goto done;
	}
	if (Irp->IoStatus.Information > data_size) {
		KdPrint(("[tdi_flt] tdi_receive_complete: invalid data size %u > %u\n",
			Irp->IoStatus.Information, data_size));
		goto done;
	}

	// run content filter

	result = content_filter(ote_conn->flt_in, data, Irp->IoStatus.Information,
		ote_conn->in_offset, &new_data, &new_data_size, tdi_alloc,
		param->ReceiveFlags & TDI_RECEIVE_EXPEDITED);

	if (result == FILTER_DENY) goto done;
	if (!new_data) {
		// data haven't been replaced
		goto done;
	}

	// we have to replace data in mdl (or in mdl chain)
	Irp->IoStatus.Information = write_mdl(Irp->MdlAddress, new_data, new_data_size);

	if (Irp->IoStatus.Information < new_data_size) {
		datapipe_t *pipe = (param->ReceiveFlags & TDI_RECEIVE_EXPEDITED) ? &ote_conn->in_oob_pipe :
			&ote_conn->in_pipe;

		status = datapipe_push(pipe, new_data + Irp->IoStatus.Information,
			new_data_size - Irp->IoStatus.Information);
		if (!status) {
			// data loss, drop connection! (for UDP just drop datagram)
			result = FILTER_DENY;
			goto done;
		}
	}

	result = FILTER_ALLOW;
done:
	if (result != FILTER_ALLOW) {
		// drop connection! (for UDP just drop datagram)

		if (ote_conn) {
			// fake disconnect! call disconnect event handler
			ot_entry_t *ote_addr = ot_find_fileobj(ote_conn->associated_fileobj, NULL);

			if (ote_addr && ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler) {
				status = ((PTDI_IND_DISCONNECT)(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler))
					(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_context,
					ote_conn->conn_ctx, 0, NULL, 0, NULL, TDI_DISCONNECT_RELEASE);
				KdPrint(("[tdi_flt] tdi_event_receive_complete: fake TDI_EVENT_DISCONNECT: 0x%x\n", status));
			}

			Irp->IoStatus.Status = STATUS_ACCESS_VIOLATION;
		}
	}
	else {
		if (param->ReceiveFlags & TDI_RECEIVE_EXPEDITED)
			ote_conn->in_oob_offset += Irp->IoStatus.Information;
		else ote_conn->in_offset += Irp->IoStatus.Information;
	}

	if (ote_conn) KeReleaseSpinLock(&g_ot_hash_guard, irql);
	if (b_free_data && data) free(data);
	if (new_data) free(new_data);

	return tdi_complete(DeviceObject, Irp, Context);
}

//----------------------------------------------------------------------------

int
tdi_receive_datagram(PIRP irp, PIO_STACK_LOCATION irps, TDI_COMPLETE_NFO *tcn)
{
	TDI_REQUEST_KERNEL_RECEIVEDG *param = (TDI_REQUEST_KERNEL_RECEIVEDG *)(&irps->Parameters);
	ot_entry_t *ote_addr = NULL;
	KIRQL irql;
	TA_ADDRESS *local_addr, *remote_addr;
	int result = FILTER_DENY, ipproto;
	cont_flt_t flt;
	ULONG data_size;

	if (irps->DeviceObject != g_udpfltobj && irps->DeviceObject != g_ipfltobj) {
		// unknown device object!
		KdPrint(("[tdi_flt] tdi_receive_datagram: unknown DeviceObject 0x%x\n", irps->DeviceObject));
		goto done;
	}
	if (irps->DeviceObject == g_udpfltobj) ipproto = IPPROTO_UDP; else ipproto = IPPROTO_IP;

	// get local address

	ote_addr = ot_find_fileobj(irps->FileObject, &irql);
	if (!ote_addr) {
		KdPrint(("[tdi_flt] tdi_receive_datagram: ot_find_fileobj(0x%x)\n", irps->FileObject));
		goto done;
	}

	KdPrint(("[tdi_flt] tdi_receive_datagram: devobj 0x%x; addrobj 0x%x (offset: %u; size: %u)\n",
		irps->DeviceObject, irps->FileObject, ote_addr->in_offset, param->ReceiveLength));

	local_addr = (TA_ADDRESS *)(ote_addr->local_addr);
	remote_addr = ((TRANSPORT_ADDRESS *)(param->ReceiveDatagramInformation->RemoteAddress))->Address;

	KdPrint(("[tdi_flt] tdi_receive_datagram: %x:%u -> %x:%u (ipproto = %d)\n",
		ntohl(((TDI_ADDRESS_IP *)(remote_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(remote_addr->Address))->sin_port),
		ntohl(((TDI_ADDRESS_IP *)(local_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(local_addr->Address))->sin_port), ipproto));

	// run quick filter

	result = quick_filter((struct sockaddr *)(&local_addr->AddressType),
		(struct sockaddr *)(&remote_addr->AddressType), sizeof(struct sockaddr_in),
		ipproto, DIRECTION_OUT, NULL, &flt);
	if (result == FILTER_DENY) goto done;

	// try to read data from datapipe
	data_size = datapipe_peek(&ote_addr->in_pipe);
	if (data_size) {
		// read data into MDL
		if (!irp->MdlAddress->Next) {
			// if unchained just read
			irp->IoStatus.Information = datapipe_get(&ote_addr->in_pipe,
				MmGetSystemAddressForMdl(irp->MdlAddress), irp->MdlAddress->ByteCount);
		}
		else {
			// if chained
			char *data = (char *)malloc(data_size);
			data_size = datapipe_get(&ote_addr->in_pipe,
				MmGetSystemAddressForMdl(irp->MdlAddress), irp->MdlAddress->ByteCount);
			irp->IoStatus.Information = write_mdl(irp->MdlAddress, data, data_size);
			free(data);
		}

		ote_addr->in_offset += irp->IoStatus.Information;

		result = FILTER_CONTENT;
		goto done;
	}

	if (!ote_addr->flt_in) {
		// don't filter - allow all
		ote_addr->in_offset += irp->IoStatus.Information;
		result = FILTER_ALLOW;
		goto done;
	}

	// set content filter in completion

	tcn->cr = tdi_receive_datagram_complete;
	tcn->context = (PVOID)flt;

	result = FILTER_ALLOW;
done:
	if (ote_addr) KeReleaseSpinLock(&g_ot_hash_guard, irql);

	return result;
}

//----------------------------------------------------------------------------

NTSTATUS
tdi_receive_datagram_complete(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context)
{
	PIO_STACK_LOCATION irps = IoGetCurrentIrpStackLocation(Irp);
	TDI_REQUEST_KERNEL_RECEIVE *param = (TDI_REQUEST_KERNEL_RECEIVE *)(&irps->Parameters);
	cont_flt_t flt = (cont_flt_t)Context;
	ULONG new_data_size, data_size;
	char *data = NULL, *new_data = NULL;
	PMDL mdl;
	ot_entry_t *ote = NULL;
	KIRQL irql;
	int result = FILTER_DENY, b_free_data = 0;
	NTSTATUS status;

	KdPrint(("[tdi_flt] tdi_receive_complete: connobj 0x%x; status 0x%x\n", irps->FileObject,
		Irp->IoStatus.Status));

	if (Irp->IoStatus.Status != STATUS_SUCCESS) {
		KdPrint(("[tdi_flt] tdi_receive_datagram_complete: status 0x%x\n", Irp->IoStatus.Status));
		goto done;
	}

	ote = ot_find_fileobj(irps->FileObject, &irql);
	if (!ote) {
		KdPrint(("[tdi_flt] tdi_receive_complete: ot_find_fileobj(0x%x)\n", irps->FileObject));
		goto done;
	}

	data = get_mdl_data(Irp->MdlAddress, &data_size, &b_free_data);
	if (!data) {
		KdPrint(("[tdi_flt] tdi_receive_complete: get_mdl_data\n"));
		goto done;
	}
	if (Irp->IoStatus.Information > data_size) {
		KdPrint(("[tdi_flt] tdi_receive_complete: invalid data size %u > %u\n",
			Irp->IoStatus.Information, data_size));
		goto done;
	}

	// run content filter

	result = content_filter(flt, data, Irp->IoStatus.Information,
		ote->in_offset, &new_data, &new_data_size, tdi_alloc,
		param->ReceiveFlags & TDI_RECEIVE_EXPEDITED);

	if (result == FILTER_DENY) goto done;
	if (!new_data) {
		// data haven't been replaced
		goto done;
	}

	// we have to replace data in mdl (or in mdl chain)
	Irp->IoStatus.Information = write_mdl(Irp->MdlAddress, new_data, new_data_size);

	if (Irp->IoStatus.Information < new_data_size) {
		datapipe_t *pipe = (param->ReceiveFlags & TDI_RECEIVE_EXPEDITED) ? &ote->in_oob_pipe :
			&ote->in_pipe;

		status = datapipe_push(pipe, new_data + Irp->IoStatus.Information,
			new_data_size - Irp->IoStatus.Information);
		if (!status) {
			// data loss, drop datagram!
			result = FILTER_DENY;
			goto done;
		}
	}

	result = FILTER_ALLOW;
done:
	if (result != FILTER_ALLOW) Irp->IoStatus.Status = STATUS_ACCESS_VIOLATION;
	else ote->in_offset += Irp->IoStatus.Information;

	if (ote) KeReleaseSpinLock(&g_ot_hash_guard, irql);
	if (b_free_data && data) free(data);
	if (new_data) free(new_data);

	return tdi_complete(DeviceObject, Irp, Context);
}

//----------------------------------------------------------------------------

int
tdi_send(PIRP irp, PIO_STACK_LOCATION irps, TDI_COMPLETE_NFO *tcn)
{
	TDI_REQUEST_KERNEL_SEND *param = (TDI_REQUEST_KERNEL_SEND *)(&irps->Parameters);
	TA_ADDRESS *local_addr, *remote_addr;
	NTSTATUS status;
	ULONG new_data_size, data_size;
	char *data = NULL, *new_data = NULL;
	PMDL mdl;
	TDI_SEND_CONTEXT *ctx = NULL;
	ot_entry_t *ote_conn = NULL;
	KIRQL irql;
	int result = FILTER_DENY, b_free_data = 0;

	if (irps->DeviceObject != g_tcpfltobj && irps->DeviceObject != g_udpfltobj) {
		// unknown device object!
		KdPrint(("[tdi_flt] tdi_send: unknown DeviceObject 0x%x\n", irps->DeviceObject));
		goto done;
	}

	// get local and remote addresses of connection

	ote_conn = ot_find_fileobj(irps->FileObject, &irql);
	if (!ote_conn) {
		KdPrint(("[tdi_flt] tdi_send: ot_find_fileobj(0x%x)\n", irps->FileObject));
		goto done;
	}

	KdPrint(("[tdi_flt] tdi_send: connobj 0x%x (offset: %u; size: %u)\n", irps->FileObject,
		ote_conn->out_offset, param->SendLength));

	local_addr = (TA_ADDRESS *)(ote_conn->local_addr);
	remote_addr = (TA_ADDRESS *)(ote_conn->remote_addr);

	KdPrint(("[tdi_flt] tdi_send: %x:%u -> %x:%u\n",
		ntohl(((TDI_ADDRESS_IP *)(local_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(local_addr->Address))->sin_port),
		ntohl(((TDI_ADDRESS_IP *)(remote_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(remote_addr->Address))->sin_port)));

	if (!ote_conn->flt_out) {
		// don't filter - allow all
		result = FILTER_ALLOW;
		goto done;
	}

	// run content filter

	data = get_mdl_data(irp->MdlAddress, &data_size, &b_free_data);
	if (!data) {
	    KdPrint(("[tdi_flt] tdi_send: get_mdl_data\n"));
		result = FILTER_DENY;
		goto done;
	}
    if (data_size < param->SendLength) {
	    KdPrint(("[tdi_flt] tdi_send: MDL chain size mismatch (%u < %u)!\n",
			data_size, param->SendLength));
		goto done;
    }

	result = content_filter(ote_conn->flt_out, data, param->SendLength,
		ote_conn->out_offset, &new_data, &new_data_size, tdi_alloc,
		param->SendFlags & TDI_SEND_EXPEDITED);

	if (result == FILTER_DENY) goto done;
	if (!new_data) {
		// data haven't been replaced
		goto done;
	}

	// we have to replace data in mdl

	ctx = (TDI_SEND_CONTEXT *)malloc(sizeof(*ctx));
	if (!ctx) {
		KdPrint(("[tdi_flt] tdi_send: malloc\n"));
		goto done;
	}

	ctx->old_mdl = irp->MdlAddress;
	ctx->buf = new_data;

    irp->MdlAddress = NULL;
    irp->MdlAddress = IoAllocateMdl(new_data, new_data_size, FALSE, FALSE, irp);
	if (!irp->MdlAddress) {
		KdPrint(("[tdi_flt] tdi_send: IoAllocateMdl\n"));
		irp->MdlAddress = ctx->old_mdl; // restore old MDL
		result = FILTER_DENY;
		goto done;
	}

    MmBuildMdlForNonPagedPool(irp->MdlAddress);

	param->SendLength = new_data_size;

	// increment offset
	ote_conn->out_offset += param->SendLength;

	tcn->cr = tdi_send_complete;
	tcn->context = ctx;

	ctx = NULL;
	new_data = NULL;

	result = FILTER_ALLOW;
done:
	if (ote_conn) {
		if (result != FILTER_ALLOW) {
			// fake disconnect! call disconnect event handler (for UDP just drop datagram)
			ot_entry_t *ote_addr = ot_find_fileobj(ote_conn->associated_fileobj, NULL);

			if (ote_addr && ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler) {
				status = ((PTDI_IND_DISCONNECT)(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler))
					(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_context,
					ote_conn->conn_ctx, 0, NULL, 0, NULL, TDI_DISCONNECT_RELEASE);
				KdPrint(("[tdi_flt] tdi_send: fake TDI_EVENT_DISCONNECT: 0x%x, conn_ctx: 0x%x\n",
					status, ote_conn->conn_ctx));
			}
		}
		else {
			if (param->SendFlags & TDI_SEND_EXPEDITED) ote_conn->out_oob_offset += param->SendLength;
			else ote_conn->out_offset += param->SendLength;
		}
		KeReleaseSpinLock(&g_ot_hash_guard, irql);
	}
	if (b_free_data && data) free(data);
	if (ctx) free(ctx);
	if (new_data) free(new_data);

	return result;
}

//----------------------------------------------------------------------------

// this function destroys replaced MDL at completion
// used not for only TDI_SEND
NTSTATUS
tdi_send_complete(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context)
{
    TDI_SEND_CONTEXT *ctx = (TDI_SEND_CONTEXT *)Context;

    // destroy replaced MDL
    IoFreeMdl(Irp->MdlAddress);
    
    // replace MDL back
    Irp->MdlAddress = ctx->old_mdl;

    // free memory
    free(ctx->buf);
    free(ctx);

	return tdi_complete(DeviceObject, Irp, Context);
}

//----------------------------------------------------------------------------

int
tdi_send_datagram(PIRP irp, PIO_STACK_LOCATION irps, TDI_COMPLETE_NFO *tcn)
{
	TDI_REQUEST_KERNEL_SENDDG *param = (TDI_REQUEST_KERNEL_SENDDG *)(&irps->Parameters);
	ot_entry_t *ote = NULL;
	KIRQL irql;
	TA_ADDRESS *local_addr, *remote_addr;
	int result = FILTER_DENY, b_free_data = 0, ipproto;
	char *new_data = NULL, *data = NULL, local_buf[TA_ADDRESS_MAX];
	TDI_SEND_CONTEXT *ctx = NULL;
	ULONG data_size, new_data_size;
	PMDL mdl;
	cont_flt_t flt;

	if (irps->DeviceObject != g_udpfltobj && irps->DeviceObject != g_ipfltobj) {
		// unknown device object!
		KdPrint(("[tdi_flt] tdi_send_datagram: unknown DeviceObject 0x%x\n", irps->DeviceObject));
		goto done;
	}
	if (irps->DeviceObject == g_udpfltobj) ipproto = IPPROTO_UDP; else ipproto = IPPROTO_IP;

	// get local address

	ote = ot_find_fileobj(irps->FileObject, &irql);
	if (!ote) {
		KdPrint(("[tdi_flt] tdi_send_datagram: ot_find_fileobj(0x%x)\n", irps->FileObject));
		goto done;
	}

	KdPrint(("[tdi_flt] tdi_send_datagram: devobj 0x%x; addrobj 0x%x (offset: %u; size: %u)\n",
		irps->DeviceObject, irps->FileObject, ote->out_offset, param->SendLength));

	local_addr = (TA_ADDRESS *)(ote->local_addr);
	remote_addr = ((TRANSPORT_ADDRESS *)(param->SendDatagramInformation->RemoteAddress))->Address;

	// run quick filter

	KdPrint(("[tdi_flt] tdi_send_datagram: %x:%u -> %x:%u (ipproto = %d)\n",
		ntohl(((TDI_ADDRESS_IP *)(local_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(local_addr->Address))->sin_port),
		ntohl(((TDI_ADDRESS_IP *)(remote_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(remote_addr->Address))->sin_port), ipproto));

	result = quick_filter((struct sockaddr *)(&local_addr->AddressType),
		(struct sockaddr *)(&remote_addr->AddressType), sizeof(struct sockaddr_in),
		ipproto, DIRECTION_OUT, NULL, &flt);

	if (result == FILTER_DENY) goto done;
	if (result == FILTER_ALLOW) {
		ote->out_offset += param->SendLength;
		goto done;
	}

	// run content filter

	data = get_mdl_data(irp->MdlAddress, &data_size, &b_free_data);
	if (!data) {
	    KdPrint(("[tdi_flt] tdi_send_datagram: get_mdl_data\n"));
		result = FILTER_DENY;
		goto done;
	}
    if (data_size < param->SendLength) {
	    KdPrint(("[tdi_flt] tdi_send_datagram: MDL chain size mismatch (%u < %u)!\n",
			data_size, param->SendLength));
		goto done;
    }

	result = content_filter(flt, data, param->SendLength, ote->out_offset,
		&new_data, &new_data_size, tdi_alloc, 0);

	if (result == FILTER_DENY) goto done;
	if (!new_data) {
		// increment offset
		ote->out_offset += param->SendLength;
		goto done;
	}

	// we have to replace data in mdl

	ctx = (TDI_SEND_CONTEXT *)malloc(sizeof(*ctx));
	if (!ctx) {
		KdPrint(("[tdi_flt] tdi_send_datagram: malloc\n"));
		goto done;
	}

	ctx->old_mdl = irp->MdlAddress;
	ctx->buf = new_data;

    irp->MdlAddress = NULL;
    irp->MdlAddress = IoAllocateMdl(new_data, new_data_size, FALSE, FALSE, irp);
	if (!irp->MdlAddress) {
		KdPrint(("[tdi_flt] tdi_send_datagram: IoAllocateMdl\n"));
		irp->MdlAddress = ctx->old_mdl; // restore old MDL
		result = FILTER_DENY;
		goto done;
	}

    MmBuildMdlForNonPagedPool(irp->MdlAddress);

	param->SendLength = new_data_size;
	ote->out_offset += param->SendLength;

	tcn->cr = tdi_send_complete;
	tcn->context = ctx;

	ctx = NULL;
	new_data = NULL;

	result = FILTER_ALLOW;
done:
	if (ote) KeReleaseSpinLock(&g_ot_hash_guard, irql);
	if (b_free_data && data) free(data);
	if (ctx) free(ctx);
	if (new_data) free(new_data);

	return result;
}

//----------------------------------------------------------------------------

int
tdi_set_event_handler(PIRP irp, PIO_STACK_LOCATION irps, TDI_COMPLETE_NFO *tcn)
{
	PTDI_REQUEST_KERNEL_SET_EVENT r = (PTDI_REQUEST_KERNEL_SET_EVENT)&irps->Parameters;
	NTSTATUS status;
	ot_entry_t *ote = NULL;
	KIRQL irql;
	int result = FILTER_DENY;
	TDI_EVENT_CONTEXT *ctx;

	KdPrint(("[tdi_flt] tdi_set_event_handler: [%s] devobj 0x%x; fileobj 0x%x; EventType: %d\n",
		r->EventHandler ? "add" : "remove", irps->DeviceObject, irps->FileObject, r->EventType));

	ote = ot_find_fileobj(irps->FileObject, &irql);
	if (!ote) {
		KdPrint(("[tdi_flt] tdi_set_event_handler: ot_find_fileobj(0x%x)\n", irps->FileObject));
		if (!r->EventHandler) result = FILTER_ALLOW; // earlier than our load fileobjects
		goto done;
	}

	if (r->EventType < 0 || r->EventType >= MAX_EVENT) {
		KdPrint(("[tdi_flt] tdi_set_event_handler: unknown EventType %d!\n", r->EventType));
		result = FILTER_ALLOW;
		goto done;
	}

	ctx = &ote->ctx[r->EventType];

	if (r->EventHandler) {
		/* add EventHandler */
		int i;

		for (i = 0; i < sizeof(g_tdi_event_handlers) / sizeof(*g_tdi_event_handlers); i++)
			if (g_tdi_event_handlers[i].event == r->EventType) break;

		if (i >= sizeof(g_tdi_event_handlers) / sizeof(*g_tdi_event_handlers)) {
			KdPrint(("[tdi_flt] tdi_set_event_handler: unknown EventType %d!\n", r->EventType));
			result = FILTER_ALLOW;
			goto done;
		}

		ctx->old_handler = r->EventHandler;
		ctx->old_context = r->EventContext;

		r->EventHandler = g_tdi_event_handlers[i].handler;
		r->EventContext = ctx;

		KdPrint(("[tdi_flt] tdi_set_event_handler: old_handler 0x%x; old_context 0x%x\n",
			r->EventHandler, r->EventContext));
	}
	else {
		/* remove EventHandler */
		ctx->old_handler = NULL;
		ctx->old_context = NULL;
	}

	result = FILTER_ALLOW;
done:
	if (ote) KeReleaseSpinLock(&g_ot_hash_guard, irql);

	return result;
}

/*[--- TDI event handlers ---]*/

NTSTATUS
tdi_event_connect(
    IN PVOID TdiEventContext,
    IN LONG RemoteAddressLength,
    IN PVOID RemoteAddress,
    IN LONG UserDataLength,
    IN PVOID UserData,
    IN LONG OptionsLength,
    IN PVOID Options,
    OUT CONNECTION_CONTEXT *ConnectionContext,
    OUT PIRP *AcceptIrp)
{
	TDI_EVENT_CONTEXT *ctx = (TDI_EVENT_CONTEXT *)TdiEventContext;
	TA_ADDRESS *remote_addr = ((TRANSPORT_ADDRESS *)RemoteAddress)->Address, *local_addr;
	ot_entry_t *ote_addr = NULL, *ote_conn = NULL;
	KIRQL irql;
	NTSTATUS status;
	int result;
	cont_flt_t flt_in, flt_out;
	PIO_STACK_LOCATION irps = NULL;

	KdPrint(("[tdi_flt] tdi_event_connect: addrobj 0x%x\n", ctx->fileobj));
	
	ote_addr = ot_find_fileobj(ctx->fileobj, &irql);
	if (!ote_addr) {
		KdPrint(("[tdi_flt] tdi_event_connect: ot_find_fileobj(0x%x)\n", ctx->fileobj));
		goto done;
	}

	local_addr = (TA_ADDRESS *)(ote_addr->local_addr);

	// run filter

	KdPrint(("[tdi_flt] tdi_event_connect: %x:%u -> %x:%u\n",
		ntohl(((TDI_ADDRESS_IP *)(remote_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(remote_addr->Address))->sin_port),
		ntohl(((TDI_ADDRESS_IP *)(local_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(local_addr->Address))->sin_port)));

	result = quick_filter((struct sockaddr *)(&remote_addr->AddressType),
		(struct sockaddr *)(&local_addr->AddressType), sizeof(struct sockaddr_in),
		IPPROTO_TCP, DIRECTION_IN, &flt_in, &flt_out);
	if (result == FILTER_DENY) goto done;

	// run handler
	status = ((PTDI_IND_CONNECT)(ctx->old_handler))
		(ctx->old_context, RemoteAddressLength, RemoteAddress,
		UserDataLength, UserData, OptionsLength, Options, ConnectionContext,
		AcceptIrp);

	if (status != STATUS_MORE_PROCESSING_REQUIRED || !*AcceptIrp) goto done;

	// create and initialize connobj

	irps = IoGetCurrentIrpStackLocation(*AcceptIrp);
	KdPrint(("[tdi_flt] tdi_event_connect: connobj 0x%x\n", irps->FileObject));

	KeReleaseSpinLock(&g_ot_hash_guard, irql); // for ot_add_fileobj
	ote_addr = NULL;

	result = FILTER_DENY;

	status = ot_add_fileobj(irps->DeviceObject, irps->FileObject, FILEOBJ_CONNOBJ, *ConnectionContext);
	if (status) {
		KdPrint(("[tdi_flt] tdi_event_connect: ot_add_fileobj: 0x%x\n", status));
		goto done;
	}

	ote_conn = ot_find_fileobj(irps->FileObject, &irql);
	if (!ote_conn) {
		KdPrint(("[tdi_flt] tdi_event_connect: ot_find_fileobj(0x%x)\n", irps->FileObject));
		goto done;
	}
	
	// associate connobj with addrobj
	ote_conn->associated_fileobj = ctx->fileobj;

	if (local_addr->AddressLength != remote_addr->AddressLength) {
		// what the ...
		KdPrint(("[tdi_flt] tdi_event_connect: different addr lengths! (%u != %u)\n",
			local_addr->AddressLength, remote_addr->AddressLength));
		goto done;
	}

	// associate remote address with connobj
	
	if (remote_addr->AddressLength > sizeof(ote_conn->remote_addr)) {
		KdPrint(("[tdi_flt] tdi_event_connect: address too long! (%u)\n", remote_addr->AddressLength));
		goto done;
	}
	memcpy(ote_conn->remote_addr, remote_addr, remote_addr->AddressLength);

	// associate local address with connobj

	if (local_addr->AddressLength > sizeof(ote_conn->local_addr)) {
		KdPrint(("[tdi_flt] tdi_event_connect: address too long! (%u)\n", local_addr->AddressLength));
		goto done;
	}
	memcpy(ote_conn->local_addr, local_addr, local_addr->AddressLength);

	ote_conn->flt_in = flt_in;
	ote_conn->flt_out = flt_out;

	result = FILTER_ALLOW;
	status = STATUS_MORE_PROCESSING_REQUIRED;

done:
	if (result == FILTER_DENY) {
		if (irps) IoFreeIrp(*AcceptIrp);

		*AcceptIrp = NULL;
		status = STATUS_CONNECTION_REFUSED;
	}

	if (ote_addr || ote_conn) KeReleaseSpinLock(&g_ot_hash_guard, irql);

	return status;
}

//----------------------------------------------------------------------------

// don't delete this "empty" event handler! we'll lost old handler address!
NTSTATUS
tdi_event_disconnect(
    IN PVOID TdiEventContext,
    IN CONNECTION_CONTEXT ConnectionContext,
    IN LONG DisconnectDataLength,
    IN PVOID DisconnectData,
    IN LONG DisconnectInformationLength,
    IN PVOID DisconnectInformation,
    IN ULONG DisconnectFlags)
{
	TDI_EVENT_CONTEXT *ctx = (TDI_EVENT_CONTEXT *)TdiEventContext;
	KIRQL irql;

	KdPrint(("[tdi_flt] tdi_event_disconnect: addrobj 0x%x; conn_ctx 0x%x\n",
		ctx->fileobj, ConnectionContext));

	return ((PTDI_IND_DISCONNECT)(ctx->old_handler))
		(ctx->old_context, ConnectionContext, DisconnectDataLength,
		DisconnectData, DisconnectInformationLength, DisconnectInformation,
		DisconnectFlags);
}

//----------------------------------------------------------------------------

NTSTATUS
tdi_event_receive(
    IN PVOID TdiEventContext,
    IN CONNECTION_CONTEXT ConnectionContext,
    IN ULONG ReceiveFlags,
    IN ULONG BytesIndicated,
    IN ULONG BytesAvailable,
    OUT ULONG *BytesTaken,
    IN PVOID Tsdu,
    OUT PIRP *IoRequestPacket)
{
	TDI_EVENT_CONTEXT *ctx = (TDI_EVENT_CONTEXT *)TdiEventContext;
	PFILE_OBJECT connobj = ot_find_conn_ctx(ctx->fileobj, ConnectionContext);
	ot_entry_t *ote_conn = NULL;
	KIRQL irql;
	TA_ADDRESS *remote_addr, *local_addr;
	int result = FILTER_DENY;
	char *new_data = NULL;
	ULONG new_data_size;
	NTSTATUS status;

	KdPrint(("[tdi_flt] tdi_event_receive: addrobj 0x%x; conn_ctx 0x%x; connobj: 0x%x\n",
		ctx->fileobj, ConnectionContext, connobj));

	ote_conn = ot_find_fileobj(connobj, &irql);
	if (!ote_conn) {
		KdPrint(("[tdi_flt] tdi_event_receive: ot_find_fileobj(0x%x)\n", connobj));
		goto done;
	}

	local_addr = (TA_ADDRESS *)(ote_conn->local_addr);
	remote_addr = (TA_ADDRESS *)(ote_conn->remote_addr);

	KdPrint(("[tdi_flt] tdi_event_receive: %x:%u -> %x:%u  (offset: %u; size: %u)\n",
		ntohl(((TDI_ADDRESS_IP *)(remote_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(remote_addr->Address))->sin_port),
		ntohl(((TDI_ADDRESS_IP *)(local_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(local_addr->Address))->sin_port),
		ote_conn->in_offset, BytesAvailable));

	if (!ote_conn->flt_in) {
		// don't filter - allow all

		// call old handler (at IRQL_DISPATCH_LEVEL)
		status = ((PTDI_IND_RECEIVE)(ctx->old_handler))
			(ctx->old_context, ConnectionContext, ReceiveFlags, BytesIndicated,
			BytesAvailable, BytesTaken, Tsdu, IoRequestPacket);

		if (status == STATUS_SUCCESS) ote_conn->in_offset += BytesAvailable;
		else if (status == STATUS_MORE_PROCESSING_REQUIRED)
			ote_conn->in_offset += *BytesTaken;

		goto done;
	}

	// run content filter

	result = content_filter(ote_conn->flt_in, Tsdu, BytesAvailable, ote_conn->in_offset,
		&new_data, &new_data_size, tdi_alloc, 0);

	if (result != FILTER_ALLOW) {
		// fake disconnect! call disconnect event handler (for UDP just drop datagram)
		ot_entry_t *ote_addr = ot_find_fileobj(ote_conn->associated_fileobj, NULL);

		if (ote_addr && ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler) {
			status = ((PTDI_IND_DISCONNECT)(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler))
				(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_context,
				ConnectionContext, 0, NULL, 0, NULL, TDI_DISCONNECT_RELEASE);
			KdPrint(("[tdi_flt] tdi_event_receive: fake TDI_EVENT_DISCONNECT: 0x%x\n", status));
	
			status = STATUS_DATA_NOT_ACCEPTED;
		}
		else {
			// push all of the data in our pipe
			status = datapipe_push(&ote_conn->in_pipe, Tsdu, BytesAvailable);
			if (!status) {
				// data loss, can't disconnect! maybe send TDI_DISCONNECT?
				status = STATUS_DATA_NOT_ACCEPTED;
			}
			else {
				*BytesTaken = BytesAvailable;
				status = STATUS_SUCCESS;
			}
		}
		goto done;
	}
	
	if (!new_data) {
		// call old handler (at IRQL_DISPATCH_LEVEL)
		status = ((PTDI_IND_RECEIVE)(ctx->old_handler))
			(ctx->old_context, ConnectionContext, ReceiveFlags, BytesIndicated,
			BytesAvailable, BytesTaken, Tsdu, IoRequestPacket);

		if (status == STATUS_SUCCESS) ote_conn->in_offset += BytesAvailable;
		else if (status == STATUS_MORE_PROCESSING_REQUIRED)
			ote_conn->in_offset += *BytesTaken;

		goto done;
	}

	// call old handler with new data (at IRQL_DISPATCH_LEVEL)
	status = ((PTDI_IND_RECEIVE)(ctx->old_handler))
		(ctx->old_context, ConnectionContext, ReceiveFlags, new_data_size,
		new_data_size, BytesTaken, new_data, IoRequestPacket);

	if (status == STATUS_SUCCESS) {
		ote_conn->in_offset += new_data_size;
		*BytesTaken = BytesAvailable;
	}
	else if (status == STATUS_MORE_PROCESSING_REQUIRED) {
		// push remaining data in our pipe
		status = datapipe_push(&ote_conn->in_pipe, new_data + *BytesTaken,
			new_data_size - *BytesTaken);
		if (!status) {
			// data loss, drop connection! (for UDP just drop datagram)
			ot_entry_t *ote_addr = ot_find_fileobj(ote_conn->associated_fileobj, NULL);

			if (ote_addr && ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler) {
				status = ((PTDI_IND_DISCONNECT)(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler))
					(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_context,
					ConnectionContext, 0, NULL, 0, NULL, TDI_DISCONNECT_RELEASE);
				KdPrint(("[tdi_flt] tdi_event_receive: fake TDI_EVENT_DISCONNECT: 0x%x\n", status));
			}
			status = STATUS_DATA_NOT_ACCEPTED;
			goto done;
		}
		ote_conn->in_offset += *BytesTaken;
	}

done:
	if (new_data) free(new_data);
	if (ote_conn) KeReleaseSpinLock(&g_ot_hash_guard, irql);

	return status;
}

//----------------------------------------------------------------------------

NTSTATUS
tdi_event_receive_datagram(
    IN PVOID TdiEventContext,
    IN LONG SourceAddressLength,
    IN PVOID SourceAddress,
    IN LONG OptionsLength,
    IN PVOID Options,
    IN ULONG ReceiveDatagramFlags,
    IN ULONG BytesIndicated,
    IN ULONG BytesAvailable,
    OUT ULONG *BytesTaken,
    IN PVOID Tsdu,
    OUT PIRP *IoRequestPacket)
{
	TDI_EVENT_CONTEXT *ctx = (TDI_EVENT_CONTEXT *)TdiEventContext;
	ot_entry_t *ote_addr = NULL;
	KIRQL irql;
	TA_ADDRESS *remote_addr, *local_addr;
	int result = FILTER_DENY;
	char *new_data = NULL;
	ULONG new_data_size;
	NTSTATUS status;
	cont_flt_t flt;

	KdPrint(("[tdi_flt] tdi_event_receive_datagram: addrobj 0x%x\n", ctx->fileobj));

	ote_addr = ot_find_fileobj(ctx->fileobj, &irql);
	if (!ote_addr) {
		KdPrint(("[tdi_flt] tdi_event_receive_datagram: ot_find_fileobj(0x%x)\n", ctx->fileobj));
		goto done;
	}

	local_addr = (TA_ADDRESS *)(ote_addr->local_addr);
	remote_addr = ((TRANSPORT_ADDRESS *)SourceAddress)->Address;

	KdPrint(("[tdi_flt] tdi_event_receive_datagram: %x:%u -> %x:%u  (offset: %u; size: %u)\n",
		ntohl(((TDI_ADDRESS_IP *)(remote_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(remote_addr->Address))->sin_port),
		ntohl(((TDI_ADDRESS_IP *)(local_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(local_addr->Address))->sin_port),
		ote_addr->in_offset, BytesAvailable));

	// run quick filter

	result = quick_filter((struct sockaddr *)(&local_addr->AddressType),
		(struct sockaddr *)(&remote_addr->AddressType), sizeof(struct sockaddr_in),
		IPPROTO_UDP, DIRECTION_OUT, NULL, &flt);

	if (result == FILTER_DENY) {
		// ignore packet
		*BytesTaken = BytesAvailable;
		status = STATUS_SUCCESS;
		goto done;
	}

	if (result == FILTER_ALLOW) {
		// call old handler (at IRQL_DISPATCH_LEVEL)
		status = ((PTDI_IND_RECEIVE_DATAGRAM)(ctx->old_handler))
			(ctx->old_context, SourceAddressLength, SourceAddress, OptionsLength,
			Options, ReceiveDatagramFlags, BytesIndicated, BytesAvailable, BytesTaken,
			Tsdu, IoRequestPacket);

		if (status == STATUS_SUCCESS) ote_addr->in_offset += BytesAvailable;
		else if (status == STATUS_MORE_PROCESSING_REQUIRED)
			ote_addr->in_offset += *BytesTaken;

		goto done;
	}

	// run content filter

	new_data = NULL;

	result = content_filter(flt, Tsdu, BytesAvailable, ote_addr->in_offset,
		&new_data, &new_data_size, tdi_alloc, 0);

	if (result != FILTER_ALLOW) {
		// ignore packet
		*BytesTaken = BytesAvailable;
		status = STATUS_SUCCESS;
		goto done;
	}
	
	if (!new_data) {
		// call old handler (at IRQL_DISPATCH_LEVEL)
		status = ((PTDI_IND_RECEIVE_DATAGRAM)(ctx->old_handler))
			(ctx->old_context, SourceAddressLength, SourceAddress, OptionsLength,
			Options, ReceiveDatagramFlags, BytesIndicated, BytesAvailable, BytesTaken,
			Tsdu, IoRequestPacket);

		if (status == STATUS_SUCCESS) ote_addr->in_offset += BytesAvailable;
		else if (status == STATUS_MORE_PROCESSING_REQUIRED)
			ote_addr->in_offset += *BytesTaken;

		goto done;
	}

	// call old handler with new data (at IRQL_DISPATCH_LEVEL)
	status = ((PTDI_IND_RECEIVE_DATAGRAM)(ctx->old_handler))
		(ctx->old_context, SourceAddressLength, SourceAddress, OptionsLength,
		Options, ReceiveDatagramFlags, new_data_size, new_data_size, BytesTaken,
		new_data, IoRequestPacket);

	if (status == STATUS_SUCCESS) ote_addr->in_offset += new_data_size;
	else if (status == STATUS_MORE_PROCESSING_REQUIRED) {
		// push remaining data in our pipe
		status = datapipe_push(&ote_addr->in_pipe, new_data + *BytesTaken,
			new_data_size - *BytesTaken);
		// if (!status) data loss, normal for datagram?
		ote_addr->in_offset += *BytesTaken;
	}

done:
	if (new_data) free(new_data);
	if (ote_addr) KeReleaseSpinLock(&g_ot_hash_guard, irql);

	return status;
}

//----------------------------------------------------------------------------

NTSTATUS
tdi_event_receive_expedited(
    IN PVOID TdiEventContext,
    IN CONNECTION_CONTEXT ConnectionContext,
    IN ULONG ReceiveFlags,
    IN ULONG BytesIndicated,
    IN ULONG BytesAvailable,
    OUT ULONG *BytesTaken,
    IN PVOID Tsdu,
    OUT PIRP *IoRequestPacket)
{
	TDI_EVENT_CONTEXT *ctx = (TDI_EVENT_CONTEXT *)TdiEventContext;
	PFILE_OBJECT connobj = ot_find_conn_ctx(ctx->fileobj, ConnectionContext);
	ot_entry_t *ote_conn = NULL;
	KIRQL irql;
	TA_ADDRESS *remote_addr, *local_addr;
	int result = FILTER_DENY;
	char *new_data = NULL;
	ULONG new_data_size;
	NTSTATUS status;

	KdPrint(("[tdi_flt] tdi_event_receive_expedited: addrobj 0x%x; conn_ctx 0x%x; connobj: 0x%x\n",
		ctx->fileobj, ConnectionContext, connobj));

	ote_conn = ot_find_fileobj(connobj, &irql);
	if (!ote_conn) {
		KdPrint(("[tdi_flt] tdi_event_receive_expedited: ot_find_fileobj(0x%x)\n", connobj));
		goto done;
	}

	local_addr = (TA_ADDRESS *)(ote_conn->local_addr);
	remote_addr = (TA_ADDRESS *)(ote_conn->remote_addr);

	KdPrint(("[tdi_flt] tdi_event_receive_expedited: %x:%u -> %x:%u  (offset: %u; size: %u)\n",
		ntohl(((TDI_ADDRESS_IP *)(remote_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(remote_addr->Address))->sin_port),
		ntohl(((TDI_ADDRESS_IP *)(local_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(local_addr->Address))->sin_port),
		ote_conn->in_oob_offset, BytesAvailable));

	if (!ote_conn->flt_in) {
		// don't filter - allow all

		// call old handler (at IRQL_DISPATCH_LEVEL)
		status = ((PTDI_IND_RECEIVE_EXPEDITED)(ctx->old_handler))
			(ctx->old_context, ConnectionContext, ReceiveFlags, BytesIndicated,
			BytesAvailable, BytesTaken, Tsdu, IoRequestPacket);

		if (status == STATUS_SUCCESS) ote_conn->in_oob_offset += BytesAvailable;
		else if (status == STATUS_MORE_PROCESSING_REQUIRED)
			ote_conn->in_oob_offset += *BytesTaken;

		goto done;
	}

	// run content filter

	new_data = NULL;

	result = content_filter(ote_conn->flt_in, Tsdu, BytesAvailable, ote_conn->in_oob_offset,
		&new_data, &new_data_size, tdi_alloc, 1);

	if (result != FILTER_ALLOW) {
		// fake disconnect! call disconnect event handler
		ot_entry_t *ote_addr = ot_find_fileobj(ote_conn->associated_fileobj, NULL);

		if (ote_addr && ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler) {
			status = ((PTDI_IND_DISCONNECT)(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler))
				(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_context,
				ConnectionContext, 0, NULL, 0, NULL, TDI_DISCONNECT_RELEASE);
			KdPrint(("[tdi_flt] tdi_event_receive_expedited: fake TDI_EVENT_DISCONNECT: 0x%x\n", status));
	
			status = STATUS_DATA_NOT_ACCEPTED;
		}
		else {
			// push all of the data in our pipe
			status = datapipe_push(&ote_conn->in_oob_pipe, Tsdu, BytesAvailable);
			if (!status) {
				// data loss, can't disconnect!
				status = STATUS_DATA_NOT_ACCEPTED;
			}
			else {
				*BytesTaken = BytesAvailable;
				status = STATUS_SUCCESS;
			}
		}
		goto done;
	}
	
	if (!new_data) {
		// call old handler (at IRQL_DISPATCH_LEVEL)
		status = ((PTDI_IND_RECEIVE_EXPEDITED)(ctx->old_handler))
			(ctx->old_context, ConnectionContext, ReceiveFlags, BytesIndicated,
			BytesAvailable, BytesTaken, Tsdu, IoRequestPacket);

		if (status == STATUS_SUCCESS) ote_conn->in_oob_offset += BytesAvailable;
		else if (status == STATUS_MORE_PROCESSING_REQUIRED)
			ote_conn->in_oob_offset += *BytesTaken;

		goto done;
	}

	// call old handler with new data (at IRQL_DISPATCH_LEVEL)
	status = ((PTDI_IND_RECEIVE_EXPEDITED)(ctx->old_handler))
		(ctx->old_context, ConnectionContext, ReceiveFlags, new_data_size,
		new_data_size, BytesTaken, new_data, IoRequestPacket);

	if (status == STATUS_SUCCESS) {
		ote_conn->in_oob_offset += new_data_size;
		*BytesTaken = BytesAvailable;
	}
	else if (status == STATUS_MORE_PROCESSING_REQUIRED) {
		// push remaining data in our pipe
		status = datapipe_push(&ote_conn->in_oob_pipe, new_data + *BytesTaken,
			new_data_size - *BytesTaken);
		if (!status) {
			// data loss, drop connection!
			ot_entry_t *ote_addr = ot_find_fileobj(ote_conn->associated_fileobj, NULL);

			if (ote_addr && ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler) {
				status = ((PTDI_IND_DISCONNECT)(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler))
					(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_context,
					ConnectionContext, 0, NULL, 0, NULL, TDI_DISCONNECT_RELEASE);
				KdPrint(("[tdi_flt] tdi_event_receive_expedited: fake TDI_EVENT_DISCONNECT: 0x%x\n", status));
			}
			status = STATUS_DATA_NOT_ACCEPTED;
			goto done;
		}
		ote_conn->in_oob_offset += *BytesTaken;
	}

done:
	if (new_data) free(new_data);
	if (ote_conn) KeReleaseSpinLock(&g_ot_hash_guard, irql);

	return status;
}

//----------------------------------------------------------------------------

NTSTATUS
tdi_event_chained_receive(
    IN PVOID TdiEventContext,
    IN CONNECTION_CONTEXT ConnectionContext,
    IN ULONG ReceiveFlags,
    IN ULONG ReceiveLength,
    IN ULONG StartingOffset,
    IN PMDL  Tsdu,
    IN PVOID TsduDescriptor)
{
	TDI_EVENT_CONTEXT *event_ctx = (TDI_EVENT_CONTEXT *)TdiEventContext;
	PFILE_OBJECT connobj = ot_find_conn_ctx(event_ctx->fileobj, ConnectionContext);
	ot_entry_t *ote_conn = NULL, *ote_addr = NULL;
	KIRQL irql;
	TA_ADDRESS *remote_addr, *local_addr;
	int result = FILTER_DENY, b_free_data = 0, b_release_mdl = 1;
	char *data = NULL, *new_data = NULL;
	ULONG data_size, new_data_size;
	PMDL mdl;
	NTSTATUS status;
	PIRP new_irp;

	KdPrint(("[tdi_flt] tdi_event_chained_receive: fileobj 0x%x; conn_ctx 0x%x; connobj: 0x%x\n",
		event_ctx->fileobj, ConnectionContext, connobj));

	ote_conn = ot_find_fileobj(connobj, &irql);
	if (!ote_conn) {
		KdPrint(("[tdi_flt] tdi_event_receive: ot_find_fileobj\n"));
		goto done;
	}

	local_addr = (TA_ADDRESS *)(ote_conn->local_addr);
	remote_addr = (TA_ADDRESS *)(ote_conn->remote_addr);

	KdPrint(("[tdi_flt] tdi_event_chained_receive: %x:%u -> %x:%u  (offset: %u; size: %u)\n",
		ntohl(((TDI_ADDRESS_IP *)(remote_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(remote_addr->Address))->sin_port),
		ntohl(((TDI_ADDRESS_IP *)(local_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(local_addr->Address))->sin_port),
		ote_conn->in_offset, ReceiveLength));

	if (!ote_conn->flt_in) {
		// don't filter - allow all

		// call old handler (at IRQL_DISPATCH_LEVEL)
		status = ((PTDI_IND_CHAINED_RECEIVE)(event_ctx->old_handler))
			(event_ctx->old_context, ConnectionContext, ReceiveFlags, ReceiveLength,
			StartingOffset, Tsdu, TsduDescriptor);

		if (status == STATUS_SUCCESS || status == STATUS_PENDING) {
			ote_conn->in_offset += ReceiveLength;
			b_release_mdl = 0;
		}

		result = FILTER_ALLOW;
		goto done;
	}

	// run content filter

	data = get_mdl_data(Tsdu, &data_size, &b_free_data);
	if (!data) {
		KdPrint(("[tdi_flt] tdi_receive_complete: get_mdl_data\n"));
		result = FILTER_DENY;
		goto done;
	}
	if (ReceiveLength > data_size) {
		KdPrint(("[tdi_flt] tdi_receive_complete: invalid data size %u > %u\n",
			ReceiveLength, data_size));
		result = FILTER_DENY;
		goto done;
	}

	result = content_filter(ote_conn->flt_in, data + StartingOffset, ReceiveLength,
		ote_conn->in_offset, &new_data, &new_data_size, tdi_alloc, 1);

	if (result != FILTER_ALLOW) goto done;
	
	if (!new_data) {
		// call old handler (at IRQL_DISPATCH_LEVEL)
		status = ((PTDI_IND_CHAINED_RECEIVE)(event_ctx->old_handler))
			(event_ctx->old_context, ConnectionContext, ReceiveFlags, ReceiveLength,
			StartingOffset, Tsdu, TsduDescriptor);

		if (status == STATUS_SUCCESS || status == STATUS_PENDING) {
			ote_conn->in_offset += ReceiveLength;
			b_release_mdl = 0;
		}

		goto done;
	}

	// call TDI_EVENT_RECEIVE handler instead of TDI_EVENT_CHAINED_RECEIVE

	ote_addr = ot_find_fileobj(event_ctx->fileobj, NULL);
	if (!ote_addr) {
		result = FILTER_DENY;
		goto done;
	}

	if (ote_addr->ctx[TDI_EVENT_RECEIVE].old_handler) {
		ULONG BytesTaken;

		new_irp = NULL;
	
		status = ((PTDI_IND_RECEIVE)(ote_addr->ctx[TDI_EVENT_RECEIVE].old_handler))
			(event_ctx->old_context, ConnectionContext,
			ReceiveFlags | TDI_RECEIVE_ENTIRE_MESSAGE,
			new_data_size, new_data_size,
			&BytesTaken, new_data, &new_irp);

		if (status == STATUS_SUCCESS) ote_conn->out_offset += new_data_size;
		else if (status == STATUS_MORE_PROCESSING_REQUIRED) {
			ote_conn->in_offset += BytesTaken;
			// push remaining data in our pipe
			status = datapipe_push(&ote_conn->in_pipe, new_data + BytesTaken,
				new_data_size - BytesTaken);
			if (!status) {
				// data loss, drop connection! (for UDP just drop datagram)
				if (ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler) {
					status = ((PTDI_IND_DISCONNECT)(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler))
						(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_context,
						ConnectionContext, 0, NULL, 0, NULL, TDI_DISCONNECT_RELEASE);
					KdPrint(("[tdi_flt] tdi_event_chained_receive: fake TDI_EVENT_DISCONNECT: 0x%x\n", status));
				}
				status = STATUS_DATA_NOT_ACCEPTED;
				goto done;
			}
			status = STATUS_SUCCESS;
		}

		if (new_irp) {
			IoCallDriver(ote_addr->devobj, new_irp);
		}
	}
	else {
		// push all of the data in our pipe
		status = datapipe_push(&ote_conn->in_pipe, data, data_size);
		if (!status) {
			// data loss, drop connection!
			if (ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler) {
				status = ((PTDI_IND_DISCONNECT)(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler))
					(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_context,
					ConnectionContext, 0, NULL, 0, NULL, TDI_DISCONNECT_RELEASE);
				KdPrint(("[tdi_flt] tdi_event_chained_receive: fake TDI_EVENT_DISCONNECT: 0x%x\n", status));
			}
			status = STATUS_DATA_NOT_ACCEPTED;
			goto done;
		}
		status = STATUS_SUCCESS;
	}

done:
	if (result != FILTER_ALLOW) {
		// fake disconnect! call disconnect event handler (for UDP just drop datagram)
		if (!ote_addr) ote_addr = ot_find_fileobj(ote_conn->associated_fileobj, NULL);

		if (ote_addr && ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler) {
			status = ((PTDI_IND_DISCONNECT)(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler))
				(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_context,
				ConnectionContext, 0, NULL, 0, NULL, TDI_DISCONNECT_RELEASE);
			KdPrint(("[tdi_flt] tdi_event_chained_receive: fake TDI_EVENT_DISCONNECT: 0x%x\n", status));
		}
		else {
			// push all data in our pipe
			status = datapipe_push(&ote_conn->in_pipe, data, data_size);
			// if (!status) data loss, can't disconnect!
		}
		status = STATUS_DATA_NOT_ACCEPTED;
	}

	if (b_release_mdl) TdiReturnChainedReceives(&TsduDescriptor, 1);
	
	if (ote_conn) KeReleaseSpinLock(&g_ot_hash_guard, irql);
	if (b_free_data && data) free(data);
	if (new_data) free(new_data);

	return status;
}

//----------------------------------------------------------------------------

NTSTATUS
tdi_event_chained_receive_datagram(
    IN PVOID TdiEventContext,
    IN LONG SourceAddressLength,
    IN PVOID SourceAddress,
    IN LONG OptionsLength,
    IN PVOID Options,
    IN ULONG ReceiveDatagramFlags,
    IN ULONG ReceiveDatagramLength,
    IN ULONG StartingOffset,
    IN PMDL  Tsdu,
    IN PVOID TsduDescriptor)
{
	TDI_EVENT_CONTEXT *event_ctx = (TDI_EVENT_CONTEXT *)TdiEventContext;
	ot_entry_t *ote_addr = NULL;
	KIRQL irql;
	TA_ADDRESS *remote_addr, *local_addr;
	int result = FILTER_DENY, b_free_data = 0, b_release_mdl = 1;
	char *data = NULL, *new_data = NULL;
	ULONG data_size, new_data_size;
	PMDL mdl;
	NTSTATUS status;
	cont_flt_t flt;
	PIRP new_irp;

	KdPrint(("[tdi_flt] tdi_event_chained_receive_datagram: addrobj 0x%x\n", event_ctx->fileobj));

	ote_addr = ot_find_fileobj(event_ctx->fileobj, &irql);
	if (!ote_addr) {
		KdPrint(("[tdi_flt] tdi_event_receive_datagram: ot_find_fileobj\n"));
		goto done;
	}

	local_addr = (TA_ADDRESS *)(ote_addr->local_addr);
	remote_addr = ((TRANSPORT_ADDRESS *)SourceAddress)->Address;

	KdPrint(("[tdi_flt] tdi_event_chained_receive_datagram: %x:%u -> %x:%u  (offset: %u; size: %u)\n",
		ntohl(((TDI_ADDRESS_IP *)(remote_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(remote_addr->Address))->sin_port),
		ntohl(((TDI_ADDRESS_IP *)(local_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(local_addr->Address))->sin_port),
		ote_addr->in_offset, ReceiveDatagramLength));

	// run quick filter

	result = quick_filter((struct sockaddr *)(&local_addr->AddressType),
		(struct sockaddr *)(&remote_addr->AddressType), sizeof(struct sockaddr_in),
		IPPROTO_UDP, DIRECTION_OUT, NULL, &flt);

	if (result == FILTER_DENY) {
		// ignore packet
		status = STATUS_SUCCESS;
		goto done;
	}

	if (result == FILTER_ALLOW) {
		// call old handler (at IRQL_DISPATCH_LEVEL)
		status = ((PTDI_IND_CHAINED_RECEIVE_DATAGRAM)(event_ctx->old_handler))
			(event_ctx->old_context, SourceAddressLength, SourceAddress, OptionsLength,
			Options, ReceiveDatagramFlags, ReceiveDatagramLength, StartingOffset,
			Tsdu, TsduDescriptor);

		if (status == STATUS_SUCCESS || status == STATUS_PENDING) {
			ote_addr->in_offset += ReceiveDatagramLength;
			b_release_mdl = 0;
		}

		result = FILTER_ALLOW;
		goto done;
	}

	// run content filter

	data = get_mdl_data(Tsdu, &data_size, &b_free_data);
	if (!data) {
		KdPrint(("[tdi_flt] tdi_receive_complete: get_mdl_data\n"));
		result = FILTER_DENY;
		goto done;
	}
	if (ReceiveDatagramLength > data_size) {
		KdPrint(("[tdi_flt] tdi_receive_complete: invalid data size %u > %u\n",
			ReceiveDatagramLength, data_size));
		result = FILTER_DENY;
		goto done;
	}

	result = content_filter(flt, data + StartingOffset, ReceiveDatagramLength,
		ote_addr->in_offset, &new_data, &new_data_size, tdi_alloc, 1);

	if (result != FILTER_ALLOW) goto done;
	
	if (!new_data) {
		// call old handler (at IRQL_DISPATCH_LEVEL)
		status = ((PTDI_IND_CHAINED_RECEIVE_DATAGRAM)(event_ctx->old_handler))
				(event_ctx->old_context, SourceAddressLength, SourceAddress, OptionsLength,
				Options, ReceiveDatagramFlags, ReceiveDatagramLength, StartingOffset,
				Tsdu, TsduDescriptor);

		if (status == STATUS_SUCCESS || status == STATUS_PENDING) {
			ote_addr->in_offset += ReceiveDatagramLength;
			b_release_mdl = 0;
		}

		goto done;
	}

	// call TDI_EVENT_RECEIVE_DATAGRAM handler instead of TDI_EVENT_CHAINED_RECEIVE_DATAGRAM

	if (ote_addr->ctx[TDI_EVENT_RECEIVE].old_handler) {
		ULONG BytesTaken;
	
		new_irp = NULL;

		status = ((PTDI_IND_RECEIVE_DATAGRAM)(ote_addr->ctx[TDI_EVENT_RECEIVE_DATAGRAM].old_handler))
			(event_ctx->old_context, SourceAddressLength, SourceAddress, OptionsLength,
			Options, ReceiveDatagramFlags | TDI_RECEIVE_ENTIRE_MESSAGE,
			new_data_size, new_data_size,
			&BytesTaken, new_data, &new_irp);

		if (status == STATUS_SUCCESS) ote_addr->out_offset += new_data_size;
		else if (status == STATUS_MORE_PROCESSING_REQUIRED) {
			ote_addr->in_offset += BytesTaken;
			// push remaining data in our pipe
			status = datapipe_push(&ote_addr->in_pipe, new_data + BytesTaken,
				new_data_size - BytesTaken);
			if (!status) {
				// data loss, drop datagram!
				status = STATUS_DATA_NOT_ACCEPTED;
				goto done;
			}
			status = STATUS_SUCCESS;
		}

		if (new_irp) {
			IoCallDriver(ote_addr->devobj, new_irp);
		}
	}
	else {
		// push all of the data in our pipe
		status = datapipe_push(&ote_addr->in_pipe, data, data_size);
		if (!status) {
			// data loss, drop datagram!
			status = STATUS_DATA_NOT_ACCEPTED;
			goto done;
		}
		status = STATUS_SUCCESS;
	}

done:
	if (result != FILTER_ALLOW) status = STATUS_DATA_NOT_ACCEPTED;

	if (b_release_mdl) TdiReturnChainedReceives(&TsduDescriptor, 1);
	
	if (ote_addr) KeReleaseSpinLock(&g_ot_hash_guard, irql);
	if (b_free_data && data) free(data);
	if (new_data) free(new_data);

	return status;
}

//----------------------------------------------------------------------------

NTSTATUS
tdi_event_chained_receive_expedited(
    IN PVOID TdiEventContext,
    IN CONNECTION_CONTEXT ConnectionContext,
    IN ULONG ReceiveFlags,
    IN ULONG ReceiveLength,
    IN ULONG StartingOffset,
    IN PMDL  Tsdu,
    IN PVOID TsduDescriptor)
{
	TDI_EVENT_CONTEXT *event_ctx = (TDI_EVENT_CONTEXT *)TdiEventContext;
	PFILE_OBJECT connobj = ot_find_conn_ctx(event_ctx->fileobj, ConnectionContext);
	ot_entry_t *ote_conn = NULL, *ote_addr = NULL;
	KIRQL irql;
	TA_ADDRESS *remote_addr, *local_addr;
	int result = FILTER_DENY, b_free_data = 0;
	char *data = NULL, *new_data = NULL;
	ULONG data_size, new_data_size;
	PMDL mdl;
	NTSTATUS status;

	KdPrint(("[tdi_flt] tdi_event_chained_receive_expedited: fileobj 0x%x; conn_ctx 0x%x; connobj: 0x%x\n",
		event_ctx->fileobj, ConnectionContext, connobj));

	ote_conn = ot_find_fileobj(connobj, &irql);
	if (!ote_conn) {
		KdPrint(("[tdi_flt] tdi_event_receive_expedited: ot_find_fileobj\n"));
		goto done;
	}

	local_addr = (TA_ADDRESS *)(ote_conn->local_addr);
	remote_addr = (TA_ADDRESS *)(ote_conn->remote_addr);

	KdPrint(("[tdi_flt] tdi_event_receive_expedited: %x:%u -> %x:%u  (offset: %u; size: %u)\n",
		ntohl(((TDI_ADDRESS_IP *)(remote_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(remote_addr->Address))->sin_port),
		ntohl(((TDI_ADDRESS_IP *)(local_addr->Address))->in_addr),
		ntohs(((TDI_ADDRESS_IP *)(local_addr->Address))->sin_port),
		ote_conn->in_oob_offset, ReceiveLength));

	if (!ote_conn->flt_in) {
		// don't filter - allow all

		// call old handler (at IRQL_DISPATCH_LEVEL)
		status = ((PTDI_IND_CHAINED_RECEIVE_EXPEDITED)(event_ctx->old_handler))
			(event_ctx->old_context, ConnectionContext, ReceiveFlags, ReceiveLength,
			StartingOffset, Tsdu, TsduDescriptor);

		if (status == STATUS_SUCCESS || status == STATUS_PENDING)
			ote_conn->in_oob_offset += ReceiveLength;

		result = FILTER_ALLOW;
		goto done;
	}

	// run content filter

	data = get_mdl_data(Tsdu, &data_size, &b_free_data);
	if (!data) {
		KdPrint(("[tdi_flt] tdi_receive_complete: get_mdl_data\n"));
		result = FILTER_DENY;
		goto done;
	}
	if (ReceiveLength > data_size) {
		KdPrint(("[tdi_flt] tdi_receive_complete: invalid data size %u > %u\n",
			ReceiveLength, data_size));
		result = FILTER_DENY;
		goto done;
	}

	result = content_filter(ote_conn->flt_in, data + StartingOffset, ReceiveLength,
		ote_conn->in_oob_offset, &new_data, &new_data_size, tdi_alloc, 1);

	if (result != FILTER_ALLOW) goto done;
	
	if (!new_data) {
		// call old handler (at IRQL_DISPATCH_LEVEL)
		status = ((PTDI_IND_CHAINED_RECEIVE_EXPEDITED)(event_ctx->old_handler))
			(event_ctx->old_context, ConnectionContext, ReceiveFlags, ReceiveLength,
			StartingOffset, Tsdu, TsduDescriptor);

		if (status == STATUS_SUCCESS || status == STATUS_PENDING)
			ote_conn->in_oob_offset += ReceiveLength;

		goto done;
	}

	// call TDI_EVENT_RECEIVE_EXPEDITED handler instead of TDI_EVENT_CHAINED_RECEIVE_EXPEDITED

	ote_addr = ot_find_fileobj(event_ctx->fileobj, NULL);
	if (!ote_addr) {
		result = FILTER_DENY;
		goto done;
	}

	if (ote_addr->ctx[TDI_EVENT_RECEIVE_EXPEDITED].old_handler) {
		ULONG BytesTaken;
	
		status = ((PTDI_IND_RECEIVE_EXPEDITED)(ote_addr->ctx[TDI_EVENT_RECEIVE_EXPEDITED].old_handler))
			(event_ctx->old_context, ConnectionContext,
			ReceiveFlags | TDI_RECEIVE_ENTIRE_MESSAGE,
			new_data_size, new_data_size,
			&BytesTaken, new_data, TsduDescriptor);

		if (status == STATUS_SUCCESS) ote_conn->out_oob_offset += new_data_size;
		else if (status == STATUS_MORE_PROCESSING_REQUIRED) {
			ote_conn->in_oob_offset += BytesTaken;
			// push remaining data in our pipe
			status = datapipe_push(&ote_conn->in_pipe, new_data + BytesTaken,
				new_data_size - BytesTaken);
			if (!status) {
				// data loss, drop connection!
				if (ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler) {
					status = ((PTDI_IND_DISCONNECT)(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler))
						(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_context,
						ConnectionContext, 0, NULL, 0, NULL, TDI_DISCONNECT_RELEASE);
					KdPrint(("[tdi_flt] tdi_event_chained_receive: fake TDI_EVENT_DISCONNECT: 0x%x\n", status));
				}
				status = STATUS_DATA_NOT_ACCEPTED;
				goto done;
			}
			status = STATUS_SUCCESS;
		}
	}
	else {
		// push all of the data in our pipe
		status = datapipe_push(&ote_conn->in_pipe, new_data, new_data_size);
		if (!status) {
			// data loss, drop connection!
			if (ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler) {
				status = ((PTDI_IND_DISCONNECT)(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler))
					(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_context,
					ConnectionContext, 0, NULL, 0, NULL, TDI_DISCONNECT_RELEASE);
				KdPrint(("[tdi_flt] tdi_event_chained_receive: fake TDI_EVENT_DISCONNECT: 0x%x\n", status));
			}
			status = STATUS_DATA_NOT_ACCEPTED;
			goto done;
		}
		status = STATUS_SUCCESS;
	}

	TdiReturnChainedReceives(&TsduDescriptor, 1);

done:
	if (result != FILTER_ALLOW) {
		// fake disconnect! call disconnect event handler
		if (!ote_addr) ote_addr = ot_find_fileobj(ote_conn->associated_fileobj, NULL);

		if (ote_addr && ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler) {
			status = ((PTDI_IND_DISCONNECT)(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_handler))
				(ote_addr->ctx[TDI_EVENT_DISCONNECT].old_context,
				ConnectionContext, 0, NULL, 0, NULL, TDI_DISCONNECT_RELEASE);
			KdPrint(("[tdi_flt] tdi_event_chained_receive_expedited: fake TDI_EVENT_DISCONNECT: 0x%x\n", status));
		}
		else {
			// push all of the data in our pipe
			status = datapipe_push(&ote_conn->in_pipe, data, data_size);
			// if (!status) data loss, drop connection!
		}
		status = STATUS_DATA_NOT_ACCEPTED;
	}
	
	if (ote_conn) KeReleaseSpinLock(&g_ot_hash_guard, irql);
	if (data) free(data);
	if (new_data) free(new_data);

	return status;
}

/*---[ utility functions ]---*/

// callback for quick_filter
void *tdi_alloc(u_long size)
{
	return malloc(size);
}

// get linear data buffer from MDL or MDL chain (may be free needed)
char *get_mdl_data(PMDL mdl_data, ULONG *size, int *b_free)
{
	PMDL mdl;
	char *data;
	ULONG data_size, data_offset;

	if (!mdl_data->Next) {
		// no chain
		*size = mdl_data->ByteCount;
		*b_free = 0;
		return MmGetSystemAddressForMdl(mdl_data);
	}

	data_size = 0;
    for (mdl = mdl_data; mdl; mdl = mdl->Next) data_size += mdl->ByteCount;

	data = (char *)malloc(data_size);
	if (!data) {
		KdPrint(("[tdi_flt] get_mdl_data: malloc\n"));
		return NULL;
	}

    data_offset = 0;
    for (mdl = mdl_data; mdl; mdl = mdl->Next) {
        memcpy(data + data_offset, MmGetSystemAddressForMdl(mdl), mdl->ByteCount);
        data_offset += mdl->ByteCount;
    }

	*size = data_size;
	*b_free = 1;

	return data;
}

// write data in MDL or in MDL chain
ULONG write_mdl(PMDL mdl_data, const char *data, ULONG size)
{
	PMDL mdl;
	ULONG data_offset;

    data_offset = 0;
    for (mdl = mdl_data; mdl; mdl = mdl->Next) {
		if (size - data_offset <= mdl->ByteCount) {
	        memcpy(MmGetSystemAddressForMdl(mdl), data + data_offset, size - data_offset);
			data_offset = size;
			break;
		}
		memcpy(MmGetSystemAddressForMdl(mdl), data + data_offset, mdl->ByteCount);
		data_offset += mdl->ByteCount;
	}

	return data_offset;
}
